from typing import Optional

from geolib.models import BaseDataClass
from geolib.models.dsheetpiling.internal import SheetPileElement
from geolib.models.dsheetpiling.settings import SheetPilingElementMaterialType


class PileProperties(BaseDataClass):
    """
    Pile selected from the model window in D-Sheet Piling

    Arguments:
     material_type : Select the material of the sheet piling element : User defined, Steel, Concrete, Wood, Synthetic.
     elastic_stiffness_ei  : Stiffness of the section per running meter.
     section_bottom_level  : the vertical co-ordinate of the bottom of the sheet piling, in relation to the reference level.
     mr_char_el : Characterictic elastic moment without safety factors.
     material_factor_gamma_m  : The partial safety factor \u03B3 m should be defined, only if the User defined material type
     is selected. Otherwise, the program will automatically apply the following factors
     (acc. to the corresponding Eurocode) to calculate the design allowable moment
     * Steel : \u03B3m = 1 , acc. to Eurocode 3 – Part 5, art. 5.1.1(4)
     * Concrete : \u03B3m = 1.1, acc. to Eurocode 2 – Part 1.1, art. 3.1.6
     * Wood : \u03B3m = 1 , acc. to Eurocode 5 – Part 1-2, art. 2.3(1)
     * Synthetic : \u03B3m = 1.2
     modification_factor_k_mod  :  The modification factor kmod should be defined, only if the User defined
     and synthetic material type is selected.
     Otherwise, the program will automatically apply the following factors
     * Steel : kmod = 1
     * Concrete : kmod = 1
     * Wood : kmod= 1
     reduction_factor_on_maximum_moment  : The reduction factor applied to the maximum allowable moment
    """

    material_type: Optional[
        SheetPilingElementMaterialType
    ] = SheetPilingElementMaterialType.Steel
    elastic_stiffness_ei: Optional[float] = None
    diameter: Optional[float] = None
    section_bottom_level: Optional[float] = None
    reduction_factor_on_ei: Optional[float] = None
    mr_char_el: Optional[float] = None
    mr_char_pl: Optional[float] = None
    modification_factor_k_mod: Optional[float] = None
    material_factor_gamma_m: Optional[float] = None
    reduction_factor_on_maximum_moment: Optional[float] = None
    note_on_reduction_factor: Optional[str] = None


class WoodenSheetPileProperties(BaseDataClass):
    """
    Wooden pile selected from the model window in D-Sheet Piling


    Arguments:
     elasticity_modulus_e_m_0_mean : The mean modulus of elasticity per running meter.
     charac_flexural_strength_f_m_0_char : The characteristic flexural strength.
     system_factor_k_sys : The system factor that considers the cooperation of structural elements.
     deform_factor_k_def : Specifies deformation factor for the hardwoods or the reduction of the stiffness caused by the saturation of softwoods.
     creep_factor_psi_2_eff : The value of considers the percentage of the load that is constantly present.
     material_factor_gamma_m : The material factor.
     modif_factor_on_f_m_0_char_short_term_k_mod_f_short : The short term modification factor of the strength to count for duration life of the wooden pile.
     modif_factor_on_f_m_0_char_long_term_k_mod_f_long : The long term modification factor of the strength to count for duration life of the wooden pile.
     modification_factor_on_e_m_0_d_k_mod_e : This value describes the modification factor of the elasticity.

    """

    elasticity_modulus_e_m_0_mean: Optional[float] = None
    charac_flexural_strength_f_m_0_char: Optional[float] = None
    system_factor_k_sys: Optional[float] = None
    deform_factor_k_def: Optional[float] = None
    creep_factor_psi_2_eff: Optional[float] = None
    material_factor_gamma_m: Optional[float] = None
    modif_factor_on_f_m_0_char_short_term_k_mod_f_short: Optional[float] = None
    modif_factor_on_f_m_0_char_long_term_k_mod_f_long: Optional[float] = None
    modification_factor_on_e_m_0_d_k_mod_e: Optional[float] = None


class SheetPileProperties(BaseDataClass):
    """
    Pile selected from the model window in D-Sheet Piling


    Arguments:
     material_type : Select the material of the sheet piling element : User defined, Steel, Concrete, Wood, Synthetic.
     acting_width : The acting width can be used when the effective width changes along the sheet piling. D-SHEET PILING uses the acting width as a multiplication factor for the sheet piling stiffness and all loads, supports and reactions, except the normal force.
     elastic_stiffness_ei : Stiffness of the section per running meter.
     section_bottom_level : the vertical co-ordinate of the bottom of the sheet piling, in relation to the reference level.
     mr_char_el: Characterictic elastic moment without safety factors.
     material_factor_gamma_m : The partial safety factor \u03B3m should be defined, only if the User defined material type is selected. Otherwise, the program will automatically apply the following factors (acc. to the corresponding Eurocode) to calculate the design allowable moment:

        * Steel: \u03B3m = 1 , acc. to Eurocode 3 – Part 5, art. 5.1.1(4)
        * Concrete: \u03B3m = 1.1, acc. to Eurocode 2 – Part 1.1, art. 3.1.6
        * Wood: \u03B3m = 1 , acc. to Eurocode 5 – Part 1-2, art. 2.3(1)
        * Synthetic: \u03B3m = 1.2

     modification_factor_k_mod :  The modification factor kmod should be defined, only if the User defined and synthetic material type is selected. Otherwise, the program will automatically apply the following factors:

        * Steel: kmod = 1
        * Concrete: kmod = 1
        * Wood: kmod= 1

     reduction_factor_on_maximum_moment : The reduction factor applied to the maximum allowable moment
     reduction_factor_on_ei : Reduction factor applied on the stiffness EI of the pile.
     coating_area : The area of coating of the sheet piling (> 1). This is defined as the
     length of the perimeter of the sheet piling section per running meter of wall [m2/m2 wall].

     height : The thickness of the sheet piling profile [mm].
     elastic_section_modulus_w_el: The section modulus (also called resisting moment in the Netherlands) of the sheet piling,
     [cm3/m], used for a Feasibility control
     section_area : The cross-sectional area of the sheet piling, [cm3/m].

    """

    material_type: Optional[
        SheetPilingElementMaterialType
    ] = SheetPilingElementMaterialType.Steel
    elastic_stiffness_ei: Optional[float] = None
    acting_width: Optional[float] = None
    section_bottom_level: Optional[
        float
    ] = None  # TODO important paramter, shouldn't be default
    height: Optional[int] = 400  # value is defined in mm
    coating_area: Optional[float] = None
    width_of_sheet_piles: Optional[float] = None
    section_area: Optional[int] = None
    elastic_section_modulus_w_el: Optional[int] = None
    reduction_factor_on_ei: Optional[float] = None
    note_on_reduction_factor: Optional[str] = None
    mr_char_el: Optional[float] = None
    mr_char_pl: Optional[float] = None
    modification_factor_k_mod: Optional[float] = None
    material_factor_gamma_m: Optional[float] = None
    reduction_factor_on_maximum_moment: Optional[float] = None


class SheetPileModelPlasticCalculationProperties(BaseDataClass):
    """
    Refering to Moment-Curvature Diagram(M-N-Kappa) in  the D-SheetPiling UI


    Arguments:
     plastic_moment_positive : The plastic moment of the positive part of the moment-curvature diagram (in compression).
     plastic_moment_negative : The plastic moment of the positive part of the moment-curvature diagram (in traction).

    .. image:: /figures/dsheetpiling/sheetpileplastic.png
        :height: 400px
        :width: 600 px
        :scale: 70 %
        :align: center
    """

    symmetrical: Optional[bool] = False
    plastic_moment_positive: Optional[float] = None
    plastic_moment_negative: Optional[float] = None


class FullPlasticCalculationProperties(BaseDataClass):
    """
    Refering to Moment-Curvature Diagram(M-N-Kappa) in  the D-SheetPiling UI


    Arguments:
     symmetrical : This option is True in case of a symmetric moment-curvature diagram.
     eI_branch_2_positive : The flexural stiffness of the 2nd branch of the moment-curvature diagram (in compression).
     eI_branch_2_negative : The flexural stiffness of the 2nd branch of the moment-curvature diagram (in traction).
     moment_point_1_positive : The limit moment of the of the 1st branch of the moment-curvature diagram (in compression).
     moment_point_1_negative : The limit moment of the of the 1st branch of the moment-curvature diagram (in traction).
     plastic_moment_positive : The plastic moment of the positive part of the moment-curvature diagram (in compression).
     plastic_moment_negative : The plastic moment of the positive part of the moment-curvature diagram (in traction).
     eI_branch_3_positive : The flexural stiffness of the 3rd branch of the moment-curvaturediagram (in compression).
     moment_point_2_positive : The limit moment of the of the 2nd branch of the moment-curvature diagram (in compression).
     eI_branch_3_negative : The flexural stiffness of the 3rd branch of the moment-curvaturediagram (in traction).
     moment_point_2_negative : The limit moment of the of the 2nd branch of the moment-curvature diagram (in traction).

    .. image:: /figures/dsheetpiling/fullplastic.png
        :height: 400px
        :width: 600 px
        :scale: 70 %
        :align: center
    """

    symmetrical: Optional[bool] = False
    eI_branch_2_positive: Optional[float] = None
    eI_branch_2_negative: Optional[float] = None
    moment_point_1_positive: Optional[float] = None
    moment_point_1_negative: Optional[float] = None
    plastic_moment_positive: Optional[float] = None
    plastic_moment_negative: Optional[float] = None
    eI_branch_3_positive: Optional[float] = None
    moment_point_2_positive: Optional[float] = None
    eI_branch_3_negative: Optional[float] = None
    moment_point_2_negative: Optional[float] = None


class DiaphragmWallProperties(BaseDataClass):
    """
    Pile selected from the model window in D-Sheet Piling

    Arguments:
        material_type : Select the material of the sheet piling element : User defined, Steel, Concrete, Wood, Synthetic.
        acting_width : The acting width can be used when the effective width changes along the sheet piling.
            D-SHEET PILING uses the acting width as a multiplication factor for the sheet piling stiffness and all loads,
            supports and reactions, except the normal force.
        elastic_stiffness_ei : Stiffness of the section per running meter.
        section_bottom_level : the vertical co-ordinate of the bottom of the sheet piling, in relation to the reference level.
        mr_char_el:: Characterictic elastic moment without safety factors.
        material_factor_gamma_m : The partial safety factor \u03B3m should be defined, only if the User defined material type
            is selected. Otherwise, the program will automatically apply the following factors
            (acc. to the corresponding Eurocode) to calculate the design allowable moment:

            * Steel: \u03B3m = 1 , acc. to Eurocode 3 – Part 5, art. 5.1.1(4) ;
            * Concrete: \u03B3m = 1.1, acc. to Eurocode 2 – Part 1.1, art. 3.1.6;
            * Wood: \u03B3m = 1 , acc. to Eurocode 5 – Part 1-2, art. 2.3(1);
            * Synthetic: \u03B3m = 1.2

        modification_factor_k_mod :  The modification factor kmod should be defined, only if the User defined
            and synthetic material type is selected.
            Otherwise, the program will automatically apply the following factors:

            * Steel: kmod = 1 ;
            * Concrete: kmod = 1 ;
            * Wood: kmod= 1 ;

        reduction_factor_on_ei : Reduction factor applied on the stiffness EI of the pile.

    """

    material_type: Optional[
        SheetPilingElementMaterialType
    ] = SheetPilingElementMaterialType.Steel
    section_bottom_level: Optional[float] = None
    elastic_stiffness_ei: Optional[float] = None
    acting_width: Optional[float] = None
    reduction_factor_on_ei: Optional[float] = None
    note_on_reduction_factor: Optional[str] = None
    mr_char_el: Optional[float] = None
    mr_char_pl: Optional[float] = None
    modification_factor_k_mod: Optional[float] = None
    material_factor_gamma_m: Optional[float] = None
    reduction_factor_on_maximum_moment: Optional[float] = None


class DiaphragmWall(BaseDataClass):
    """
    Diaphragm Wall selected from the model window in D-Sheet Piling


    Arguments:
     name : Name of the element inputted
     diaphragm_wall_properties : properties specifically used for the Diaphragm Wall element.
     plastic_properties : plastic calculation properties. For the Diaphragm Wall option, these are always present.
    """

    name: str = ""
    diaphragm_wall_properties: Optional[
        DiaphragmWallProperties
    ] = DiaphragmWallProperties()
    plastic_properties: Optional[
        FullPlasticCalculationProperties
    ] = FullPlasticCalculationProperties()

    def to_internal(self) -> SheetPileElement:
        return SheetPileElement(
            name=self.name,
            sheetpilingelementmaterialtype=self.diaphragm_wall_properties.material_type,
            sheetpilingelementei=self.diaphragm_wall_properties.elastic_stiffness_ei,
            sheetpilingelementwidth=self.diaphragm_wall_properties.acting_width,
            sheetpilingelementlevel=self.diaphragm_wall_properties.section_bottom_level,
            sheetpilingelementreductionfactorei=self.diaphragm_wall_properties.reduction_factor_on_ei,
            sheetpilingelementnote=self.diaphragm_wall_properties.note_on_reduction_factor,
            sheetpilingelementmaxcharacteristicmoment=self.diaphragm_wall_properties.mr_char_el,
            sheetpilingelementmaxplasticcharacteristicmoment=self.diaphragm_wall_properties.mr_char_pl,
            sheetpilingelementkmod=self.diaphragm_wall_properties.modification_factor_k_mod,
            sheetpilingelementmaterialfactor=self.diaphragm_wall_properties.material_factor_gamma_m,
            ssheetpilingelementreductionfactormaxmoment=self.diaphragm_wall_properties.reduction_factor_on_maximum_moment,
            diaphragmwallissymmetric=self.plastic_properties.symmetrical,
            diaphragmwallposeielastoplastic1=self.plastic_properties.eI_branch_2_positive,
            diaphragmwallnegeielastoplastic1=self.plastic_properties.eI_branch_2_negative,
            diaphragmwallposmomelastic=self.plastic_properties.moment_point_1_positive,
            diaphragmwallnegmomelastic=self.plastic_properties.moment_point_1_negative,
            diaphragmwallposmomplastic=self.plastic_properties.plastic_moment_positive,
            diaphragmwallnegmomplastic=self.plastic_properties.plastic_moment_negative,
            diaphragmwallposeielastoplastic2=self.plastic_properties.eI_branch_3_positive,
            diaphragmwallposmomelastoplastic=self.plastic_properties.moment_point_2_positive,
            diaphragmwallnegeielastoplastic2=self.plastic_properties.eI_branch_3_negative,
            diaphragmwallnegmomelastoplastic=self.plastic_properties.moment_point_2_negative,
        )


class Sheet(BaseDataClass):
    """
    Sheet Piling selected from the model window in D-Sheet Piling

    Arguments:
     name : Name of the element inputted
     sheet_pile_properties : properties specifically used for the Sheet pile element.
     plastic_properties : plastic calculation properties. If Elastic calculation checkbox is not selected.
    """

    name: str = ""
    sheet_pile_properties: Optional[SheetPileProperties] = SheetPileProperties()
    plastic_properties: Optional[
        SheetPileModelPlasticCalculationProperties
    ] = SheetPileModelPlasticCalculationProperties()
    wooden_sheet_pile_properties: Optional[
        WoodenSheetPileProperties
    ] = WoodenSheetPileProperties()

    def to_internal(self) -> SheetPileElement:
        return SheetPileElement(
            name=self.name,
            sheetpilingelementmaterialtype=self.sheet_pile_properties.material_type,
            sheetpilingelementei=self.sheet_pile_properties.elastic_stiffness_ei,
            sheetpilingelementwidth=self.sheet_pile_properties.acting_width,
            sheetpilingelementlevel=self.sheet_pile_properties.section_bottom_level,
            sheetpilingelementheight=self.sheet_pile_properties.height,
            sheetpilingelementcoatingarea=self.sheet_pile_properties.coating_area,
            sheetpilingpilewidth=self.sheet_pile_properties.width_of_sheet_piles,
            sheetpilingelementsectionarea=self.sheet_pile_properties.section_area,
            sheetpilingelementresistingmoment=self.sheet_pile_properties.elastic_section_modulus_w_el,
            sheetpilingelementreductionfactorei=self.sheet_pile_properties.reduction_factor_on_ei,
            sheetpilingelementnote=self.sheet_pile_properties.note_on_reduction_factor,
            sheetpilingelementmaxcharacteristicmoment=self.sheet_pile_properties.mr_char_el,
            sheetpilingelementmaxplasticcharacteristicmoment=self.sheet_pile_properties.mr_char_pl,
            sheetpilingelementkmod=self.sheet_pile_properties.modification_factor_k_mod,
            sheetpilingelementmaterialfactor=self.sheet_pile_properties.material_factor_gamma_m,
            ssheetpilingelementreductionfactormaxmoment=self.sheet_pile_properties.reduction_factor_on_maximum_moment,
            diaphragmwallissymmetric=self.plastic_properties.symmetrical,
            diaphragmwallposmomplastic=self.plastic_properties.plastic_moment_positive,
            diaphragmwallnegmomplastic=self.plastic_properties.plastic_moment_negative,
            woodensheetpilingelemente=self.wooden_sheet_pile_properties.elasticity_modulus_e_m_0_mean,
            woodensheetpilingelementcharacflexuralstrength=self.wooden_sheet_pile_properties.charac_flexural_strength_f_m_0_char,
            woodensheetpilingelementksys=self.wooden_sheet_pile_properties.system_factor_k_sys,
            woodensheetpilingelementkdef=self.wooden_sheet_pile_properties.deform_factor_k_def,
            woodensheetpilingelementpsi2eff=self.wooden_sheet_pile_properties.creep_factor_psi_2_eff,
            woodensheetpilingelementmaterialfactor=self.wooden_sheet_pile_properties.material_factor_gamma_m,
            woodensheetpilingelementkmodfshort=self.wooden_sheet_pile_properties.modif_factor_on_f_m_0_char_short_term_k_mod_f_short,
            woodensheetpilingelementkmodflong=self.wooden_sheet_pile_properties.modif_factor_on_f_m_0_char_long_term_k_mod_f_long,
            woodensheetpilingelementkmode=self.wooden_sheet_pile_properties.modification_factor_on_e_m_0_d_k_mod_e,
        )


class Pile(BaseDataClass):
    """
    Pile selected from the model window in D-Sheet Piling

    Arguments:
     name: Name of the element inputted
     pile_properties: properties specifically used for the Pile element.
     plastic_properties: plastic calculation properties. If Elastic calculation checkbox is not selected.
    """

    name: str = ""
    pile_properties: Optional[PileProperties] = PileProperties()
    plastic_properties: Optional[
        FullPlasticCalculationProperties
    ] = FullPlasticCalculationProperties()

    def to_internal(self) -> SheetPileElement:
        return SheetPileElement(
            name=self.name,
            sheetpilingelementmaterialtype=self.pile_properties.material_type,
            sheetpilingelementei=self.pile_properties.elastic_stiffness_ei,
            sheetpilingelementwidth=self.pile_properties.diameter,
            sheetpilingelementlevel=self.pile_properties.section_bottom_level,
            sheetpilingelementreductionfactorei=self.pile_properties.reduction_factor_on_ei,
            sheetpilingelementnote=self.pile_properties.note_on_reduction_factor,
            sheetpilingelementmaxcharacteristicmoment=self.pile_properties.mr_char_el,
            sheetpilingelementmaxplasticcharacteristicmoment=self.pile_properties.mr_char_pl,
            sheetpilingelementkmod=self.pile_properties.modification_factor_k_mod,
            sheetpilingelementmaterialfactor=self.pile_properties.material_factor_gamma_m,
            ssheetpilingelementreductionfactormaxmoment=self.pile_properties.reduction_factor_on_maximum_moment,
            diaphragmwallissymmetric=self.plastic_properties.symmetrical,
            diaphragmwallposeielastoplastic1=self.plastic_properties.eI_branch_2_positive,
            diaphragmwallnegeielastoplastic1=self.plastic_properties.eI_branch_2_negative,
            diaphragmwallposmomelastic=self.plastic_properties.moment_point_1_positive,
            diaphragmwallnegmomelastic=self.plastic_properties.moment_point_1_negative,
            diaphragmwallposmomplastic=self.plastic_properties.plastic_moment_positive,
            diaphragmwallnegmomplastic=self.plastic_properties.plastic_moment_negative,
            diaphragmwallposeielastoplastic2=self.plastic_properties.eI_branch_3_positive,
            diaphragmwallposmomelastoplastic=self.plastic_properties.moment_point_2_positive,
            diaphragmwallnegeielastoplastic2=self.plastic_properties.eI_branch_3_negative,
            diaphragmwallnegmomelastoplastic=self.plastic_properties.moment_point_2_negative,
        )



import logging
import re
from typing import Any, Dict, List, Tuple, get_type_hints

from geolib.errors import ParserError
from geolib.models.dseries_parser import (
    DSerieListStructure,
    DSerieParser,
    DSeriesInlineProperties,
    DSeriesRepeatedGroupsWithInlineMappedProperties,
    DSeriesStructure,
    DSeriesUnmappedNameProperties,
    DSeriesWrappedTableStructure,
    get_line_property_value,
)

logger = logging.getLogger(__name__)


class DSeriesPilingDumpParserStructure(DSeriesStructure):
    @classmethod
    def parse_text(cls, text: str):
        if "[ECHO OF MSHEET INPUT]" in text:
            logger.warning("Replacing headers to create [DUMPFILE]")
            text = text.replace("[ECHO OF MSHEET INPUT]", "[DUMPFILE]").replace(
                "[END OF DUMP]", "[END OF OUTPUT DATA]\n[END OF DUMPFILE]"
            )
        if "[END OF DUMPFILE]" not in text:
            text += "\n[END OF OUTPUT DATA]\n[END OF DUMPFILE]\n"
        return super().parse_text(text)


class DSeriesPilingParserStructure(DSeriesStructure):
    @classmethod
    def parse_text(cls, text: str):
        if "[INPUT DATA]" not in text:
            logger.warning("Putting [INPUT DATA]")
            text = text.replace("[SOIL COLLECTION]", "[INPUT DATA]\n[SOIL COLLECTION]")
        if "[END OF INPUT FILE]" in text:
            text = text.replace("[END OF INPUT FILE]", "[END OF INPUT DATA]")
        if "[End of Input Data]" in text:
            logger.warning("Putting [OUTPUT DATA]")
            text = text.replace(
                "[End of Input Data]", "[End of Input Data]\n[OUTPUT DATA]"
            )

        return super().parse_text(text)


class DSheetpilingSurchargeLoad(DSeriesRepeatedGroupsWithInlineMappedProperties):
    @classmethod
    def get_list_field_names(cls) -> List[str]:
        super_field_names = super().get_list_field_names()
        super_field_names.append("point")
        return super_field_names

    @classmethod
    def get_validated_mappings(cls, generated_dict: Dict[str, str]) -> Dict[str, str]:
        name_field = generated_dict.pop("", "")
        # Verify the name field has not been replaced by "surcharge_load"
        if "surcharge_load" in generated_dict:
            name_field = generated_dict.pop("surcharge_load")
        generated_dict["name"] = name_field
        generated_dict["points"] = generated_dict.pop("point", [])
        return generated_dict


class DSheetpilingWithNumberOfRowsTable(DSeriesWrappedTableStructure):
    @classmethod
    def parse_text(cls, text):
        """These sort of tables contain the number of lines in the first row
        the rest of its structure is similar to DSeriesWrappedTableStructure.

        Args:
            text (str): Full contains containing table structure and values.

        Returns:
            DSeriesStructure: Fully parsed structure.
        """
        splitted_text = text.split("\n")
        # Skip validating for now.
        splitted_text.pop(0)
        return super().parse_text("\n".join(splitted_text))


class DSheetpilingUnwrappedTable(DSeriesStructure):
    @classmethod
    def parse_text(cls, text):
        lines = text.split("\n")
        nrow = int(lines.pop(0).split()[0])
        # Remove the line with the column names.
        lines.pop(0)
        # Count the rest of the lines.
        if not nrow == len(lines):
            raise ParserError(
                f"Error parsing for {cls}, header indicates {nrow} lines, while there are {len(lines)} lines."
            )

        return cls(**{cls.__name__.lower(): lines})


class DSheetpilingTableEntry(DSeriesStructure):
    """Parses a table entry where the latest column is actually a name (which can be composed by a few strings.)
    E.g.:
        Nr        Level        E-mod     Cross sect.    Length     YieldF   Side
        1 -10.00  2.100E+0008  1.000E-0004  10.00 500.00 -30  0.00 2 Strut
    """

    @classmethod
    def parse_text(cls, text):
        values = [value for value in text.split() if value]
        # Remove the first value (it's just the line number)
        values.pop(0)

        # In later versions __slots__ is a type hint too, function below is to ignore this key
        type_hints = get_type_hints(cls)
        if "__slots__" in type_hints:
            type_hints.pop("__slots__")

        number_of_fields = len(type_hints)
        name_field = " ".join(
            name_value for name_value in values[number_of_fields - 1 :] if name_value
        )
        rest_fields = values[0 : number_of_fields - 1]
        rest_fields.insert(0, name_field)

        return cls(**dict(zip(type_hints.keys(), rest_fields)))



"""
This module handles the three types of reinforcements in DStability.
"""

import abc
from typing import List, Tuple

from pydantic import NoneStr, confloat

from geolib.models import BaseDataClass

from ...geometry.one import Point
from ...utils import snake_to_camel
from .internal import (
    PersistableForbiddenLine,
    PersistableGeotextile,
    PersistableNail,
    PersistablePoint,
    PersistableStressAtDistance,
)


class DStabilityReinforcement(BaseDataClass, metaclass=abc.ABCMeta):
    """Base Class for Reinforcements."""

    label: NoneStr = None

    @abc.abstractmethod
    def _to_internal_datastructure(self):
        raise NotImplementedError


class Nail(DStabilityReinforcement):
    """DStability Nail, used for soil nailing."""

    location: Point  # TODO x, z must be below ground level
    direction: float = 0.0
    horizontal_spacing: float = 0.0
    length: confloat(gt=0) = 3.0
    diameter: confloat(gt=0) = 0.1
    grout_diameter: confloat(gt=0) = 0.1
    critical_angle: float = 0.0
    max_pull_force: float = 0.0
    plastic_moment: float = 0.0
    bending_stiffness: confloat(gt=0) = 0.1
    use_facing: bool = False
    use_lateral_stress: bool = (
        False  # TODO set on wether or not lateralstresses are provided?
    )
    use_shear_stress: bool = (
        False  # TODO set on wether or not shearstresses are provided?
    )
    lateral_stresses: List[Tuple[float, float]] = []
    shear_stresses: List[Tuple[float, float]] = []

    def _to_internal_datastructure(self) -> PersistableNail:
        data = {
            **{
                snake_to_camel(name): value
                for name, value in self.dict().items()
                if name not in {"x", "z", "lateral_stresses", "shear_stresses"}
            },
            "Location": PersistablePoint(X=self.location.x, Z=self.location.z),
            "LateralStresses": [
                PersistableStressAtDistance(Distance=distance, Stress=stress)
                for distance, stress in self.lateral_stresses
            ],
            "ShearStresses": [
                PersistableStressAtDistance(Distance=distance, Stress=stress)
                for distance, stress in self.shear_stresses
            ],
        }
        return PersistableNail(**data)


class ForbiddenLine(DStabilityReinforcement):
    """Inherits :class:`~geolib.models.dstability.reinforcements.DStabilityReinforcement`. Needs to be further defined."""

    start: Point
    end: Point

    def _to_internal_datastructure(self) -> PersistableForbiddenLine:
        return PersistableForbiddenLine(
            Label=self.label,
            Start=PersistablePoint(X=self.start.x, Z=self.start.z),
            End=PersistablePoint(X=self.end.x, Z=self.end.z),
        )


class Geotextile(DStabilityReinforcement):
    """Geotextile"""

    start: Point
    end: Point
    effective_tensile_strength: confloat(gt=0)
    reduction_area: confloat(gt=0)

    def _to_internal_datastructure(self) -> PersistableForbiddenLine:
        return PersistableGeotextile(
            Label=self.label,
            Start=PersistablePoint(X=self.start.x, Z=self.start.z),
            End=PersistablePoint(X=self.end.x, Z=self.end.z),
            TensileStrength=self.effective_tensile_strength,
            ReductionArea=self.reduction_area,
        )



import logging
import warnings
from abc import ABCMeta
from enum import Enum, IntEnum
from inspect import cleandoc
from typing import Any, Dict, List, Optional, Set, Tuple, Union

from pydantic import confloat, conint, conlist, constr

import geolib.models.dsheetpiling.constructions as constructions
from geolib.geometry import Point
from geolib.models import BaseDataClass
from geolib.models.dseries_parser import (
    DSerieListStructure,
    DSeriesInlineMappedProperties,
    DSeriesInlineReversedProperties,
    DSeriesNoParseSubStructure,
    DSeriesRepeatedGroupedProperties,
    DSeriesStructure,
    DSeriesStructureCollection,
    DSeriesTableStructure,
    DSeriesUnmappedNameProperties,
    DSeriesWrappedTableStructure,
    DSerieVersion,
)
from geolib.utils import make_newline_validator

from .calculation_options import (
    CalculationOptionsPerStage as CalculationOptionsPerStageExternal,
)
from .dsheetpiling_structures import (
    DSeriesPilingDumpParserStructure,
    DSeriesPilingParserStructure,
    DSheetpilingSurchargeLoad,
    DSheetpilingTableEntry,
    DSheetpilingUnwrappedTable,
    DSheetpilingWithNumberOfRowsTable,
)
from .dsheetpiling_validator import DSheetPilingValidator
from .internal_partial_factors import (
    PartialFactorsCurI,
    PartialFactorsCurIi,
    PartialFactorsCurIii,
    PartialFactorsEc7BSet1,
    PartialFactorsEc7BSet2,
    PartialFactorsEc7Nl0,
    PartialFactorsEc7Nl1,
    PartialFactorsEc7Nl2,
    PartialFactorsEc7Nl3,
    PartialFactorsEc7SeVk1,
    PartialFactorsEc7SeVk2,
    PartialFactorsEc7SeVk3,
    PartialFactorsEurocodeDa1Set1,
    PartialFactorsEurocodeDa1Set2,
    PartialFactorsEurocodeDa2,
    PartialFactorsEurocodeDa3,
)
from .settings import (
    CalculationType,
    CurveSettings,
    DesignType,
    DistributionType,
    EarthPressureCoefficients,
    GrainType,
    HorizontalBehaviorType,
    LambdaType,
    LateralEarthPressureMethod,
    LateralEarthPressureMethodStage,
    LoadTypeFavourableUnfavourable,
    LoadTypeFavourableUnfavourableMoment,
    LoadTypePermanentVariable,
    ModelType,
    ModulusReactionType,
    ModulusSubgradeReaction,
    PartialFactorCalculationType,
    PartialFactorSetCUR,
    PartialFactorSetEC,
    PartialFactorSetEC7NADB,
    PartialFactorSetEC7NADNL,
    PartialFactorSetVerifyEC,
    PassiveSide,
    SheetPilingElementMaterialType,
    Side,
    SinglePileLoadOptions,
    SoilTypeModulusSubgradeReaction,
    VerifyType,
)

logger = logging.getLogger(__name__)


_DEFAULT_WATER_LEVEL_NAME: str = "New Water Level"
_DEFAULT_SOIL_PROFILE_NAME: str = "New Profile"
_DEFAULT_SURFACE_NAME: str = "New Surface"
_DEFAULT_SHEET_PILING_ELEMENT_NAME: str = "New element"
_DEFAULT_UNIT_WEIGHT_WATER: float = 9.81
_DEFAULT_PRE_STRESS: float = 0.0

_DEFAULT_SOIL_VERSION: int = 1007
_DEFAULT_SHEETPILING_VERSION: int = 1026

REQ_RUN_LINES = 2


class Model(DSeriesInlineReversedProperties):
    model: ModelType = ModelType.SHEET_PILING
    method: LateralEarthPressureMethod = LateralEarthPressureMethod.MIXED
    verification: bool = False
    ec7se_visible: bool = False
    pile_load_option: bool = False
    pile_load_by_user: bool = False
    probabilistic: bool = False
    check_vertical_balance: bool = True
    trildens_calculation: bool = True
    diepwand_calculation: bool = False
    elastic_calculation: bool = True
    wooden_sheetpiling: bool = False


class Version(DSerieVersion):
    soil: int = _DEFAULT_SOIL_VERSION
    d__sheet_piling: int = _DEFAULT_SHEETPILING_VERSION


class VersionExternals(DSeriesInlineMappedProperties):
    dgscptipcalc____dll: str = "19.1.1.23469"
    dgscptipui____dll: str = "19.1.1.23469"


class Soil(DSeriesUnmappedNameProperties):
    name: constr(min_length=1, max_length=25) = "default soil"
    soilcolor: int = 9764853
    soilsoiltype: SoilTypeModulusSubgradeReaction = SoilTypeModulusSubgradeReaction.SAND
    soilgraintype: GrainType = GrainType.FINE
    soilgamdry: confloat(ge=0, le=100) = 0.01
    soilgamwet: confloat(ge=0, le=100) = 0.01
    soilrelativedensity: confloat(ge=0, le=100) = 0
    soilemodmenard: confloat(ge=0.01, le=1000000) = 0.01
    soilcohesion: confloat(ge=0, le=1000000000) = 0
    soilphi: confloat(ge=-89, le=89) = 0
    soildelta: confloat(ge=-80, le=89) = 0.00
    soilcutop: float = 0  # fixed value
    soilcubottom: float = 0  # fixed value
    soilcubearingcapacityfactor: float = 4.10  # fixed value
    soilocr: confloat(ge=0, le=1000) = 1.00
    soildrained: int = 0  # fixed value
    soilpermeabkx: confloat(ge=0.00000000001, le=1000000) = 0.0001
    soilstdcohesion: confloat(ge=0, le=100000000) = 0.00
    soilstdphi: confloat(ge=0, le=100000000) = 0.00
    soildistcohesion: DistributionType = DistributionType.NONE
    soildistphi: DistributionType = DistributionType.NONE
    soilla: confloat(ge=0, le=1000) = 0
    soilln: confloat(ge=0, le=1000) = 0.01
    soillp: confloat(ge=0, le=1000) = 0.01
    soilusemenard: ModulusSubgradeReaction = ModulusSubgradeReaction.MANUAL
    soilusebrinchhansen: EarthPressureCoefficients = EarthPressureCoefficients.MANUAL
    soilshellfactor: confloat(ge=1, le=1000) = 1.00
    soillambdatype: LambdaType = LambdaType.MANUAL
    soillam1: confloat(ge=0, le=100) = 50
    soillam2: confloat(ge=0, le=100) = 80
    soillam3: confloat(ge=0, le=100) = 90
    soilkb0: confloat(ge=0.01, le=10000000000) = 1.00000e04
    soilkb1: confloat(ge=0.01, le=10000000000) = 5.00000e00
    soilkb2: confloat(ge=0.01, le=10000000000) = 1.00000e04
    soilkb3: confloat(ge=0.01, le=10000000000) = 1.00000e04
    soilkb4: confloat(ge=0.01, le=10000000000) = 1.00000e04
    soilko0: confloat(ge=0.01, le=10000000000) = 1.00000e04
    soilko1: confloat(ge=0.01, le=10000000000) = 5.00000e00
    soilko2: confloat(ge=0.01, le=10000000000) = 1.00000e04
    soilko3: confloat(ge=0.01, le=10000000000) = 1.00000e04
    soilko4: confloat(ge=0.01, le=10000000000) = 1.00000e04
    soilcurkb1: confloat(ge=0.01, le=10000000000) = 2.00000e03
    soilcurkb2: confloat(ge=0.01, le=10000000000) = 8.00000e02
    soilcurkb3: confloat(ge=0.01, le=10000000000) = 5.00000e02
    soilcurko1: confloat(ge=0.01, le=10000000000) = 2.00000e03
    soilcurko2: confloat(ge=0.01, le=10000000000) = 8.00000e02
    soilcurko3: confloat(ge=0.01, le=10000000000) = 5.00000e02
    soilhorizontalbehaviourtype: HorizontalBehaviorType = HorizontalBehaviorType.ELASTIC
    soilelasticity: float = 1.00000e03  # fixed value
    soildefaultelasticity: int = 1  # fixed value


class SoilCollection(DSeriesStructureCollection):
    soil: List[Soil] = []

    @property
    def soil_names(self) -> Set[str]:
        return {soil.name for soil in self.soil}

    def add_soil_if_unique(self, soil) -> None:
        for added_soil in self.soil:
            if soil.name == added_soil.name:
                raise NameError(f"Soil with name {soil.name} already exists.")
        self.soil.append(soil)
        return soil

    def find_soil_id(self, key) -> int:
        for i, soil in enumerate(self.soil):
            if soil.name == key:
                return i
        raise KeyError(f"Soil with {key} not present.")

    def __getitem__(self, key) -> Soil:
        if isinstance(key, int):
            return self.soil[key]
        elif isinstance(key, str):
            for soil in self.soil:
                if soil.name == key:
                    return soil
            raise KeyError(key)


class SoilLayer(DSeriesNoParseSubStructure):
    top_of_layer: float
    soil: str
    water_pressure_top: float
    water_pressure_bottom: float


class SoilProfile(DSeriesNoParseSubStructure):
    name: constr(min_length=1, max_length=25)
    layers: List[SoilLayer]
    coordinate: Point


class SoilProfiles(DSeriesNoParseSubStructure):
    soil_profiles: List[SoilProfile] = []
    curve_number: conint(ge=1, le=4) = 3
    use_unloading_reloading_curve: bool = False
    modulus_reaction_type: int = ModulusReactionType.SECANT.value

    @property
    def soil_profile_names(self) -> Set[str]:
        return {soil_profile.name for soil_profile in self.soil_profiles}

    @property
    def referenced_soil_names(self) -> Set[str]:
        return {layer.soil for profile in self.soil_profiles for layer in profile.layers}


class CalculationOptions(DSeriesStructure):
    # TODO these first values are not used for single pile
    calcfirststageinitial: bool = False
    calcminornodeson: bool = False
    calcreducedeltas: bool = False
    calcempiricalfactorstresstype: int = 0  # fixed value
    inputcalculationtype: CalculationType = CalculationType.STANDARD
    isvibrationcalculation: bool = False
    allowableanchorforcecalculationtype: bool = False
    # available when c, phi , delta is selected for as a model
    calcautolambdason: bool = True
    # design sheet pile length calculation
    designstage: conint(ge=0) = 0
    designpilelengthfrom: confloat(ge=1, le=100) = 1
    designpilelengthto: confloat(ge=1, le=100) = 1
    designpilelengthdecrement: confloat(ge=0.01, le=10) = 0.01
    designpilelengthnew: int = 1  # fixed value
    designtype: DesignType = DesignType.REPRESENTATIVE
    designeurocodepartialfactorset: PartialFactorSetEC = PartialFactorSetEC.DA1SET1
    designpartialfactorsetec7nadnl: PartialFactorSetEC7NADNL = (
        PartialFactorSetEC7NADNL.RC0
    )
    designec7nlmethod: PartialFactorCalculationType = PartialFactorCalculationType.METHODA
    designpartialfactorsetec7nadb: PartialFactorSetEC7NADB = PartialFactorSetEC7NADB.SET1
    designec7bmethod: PartialFactorCalculationType = PartialFactorCalculationType.METHODA
    designpartialfactorset: PartialFactorSetCUR = PartialFactorSetCUR.CLASSI
    designcurmethod: PartialFactorCalculationType = PartialFactorCalculationType.METHODA
    designpartialfactorsetec7nadse: int = 0  # fixed value
    designec7semethod: int = 0  # fixed value
    # verify sheet piling calculation
    verifytype: VerifyType = VerifyType.CUR
    eurocodepartialfactorset: PartialFactorSetVerifyEC = PartialFactorSetVerifyEC.DA1
    eurocodeoverallstability: bool = False
    ec7nlmethod: PartialFactorCalculationType = PartialFactorCalculationType.METHODA
    ec7nloverallpartialfactorset: PartialFactorSetEC7NADNL = PartialFactorSetEC7NADNL.RC0
    ec7nloverallanchorfactor: confloat(ge=0.001, le=1000) = 1
    ec7nadnloverallstability: bool = False
    ec7boverallstability: bool = False
    nbmethod: PartialFactorCalculationType = PartialFactorCalculationType.METHODA
    curmethod: PartialFactorCalculationType = PartialFactorCalculationType.METHODA
    curoverallpartialfactorset: PartialFactorSetCUR = PartialFactorSetCUR.CLASSI
    curoverallanchorfactor: confloat(ge=0.001, le=1000) = 1
    curoverallstability: bool = False
    ec7semethod: int = 0  # fixed value
    ec7seoverallpartialfactorset: int = 0  # fixed value
    ec7nadseoverallstability: int = 0  # fixed value
    # Characteristic Kranz Anchor Strength calculation
    curanchorforcestage: conint(ge=0) = 0
    # Overall stability calculation
    curstabilitystage: conint(ge=0) = 0
    overallstabilitytype: DesignType = DesignType.REPRESENTATIVE
    stabilityeurocodepartialfactorset: PartialFactorSetEC = PartialFactorSetEC.DA1SET1
    stabilityec7nlpartialfactorset: PartialFactorSetEC7NADNL = (
        PartialFactorSetEC7NADNL.RC0
    )
    stabilityec7bpartialfactorset: PartialFactorSetEC7NADB = PartialFactorSetEC7NADB.SET1
    stabilitycurpartialfactorset: PartialFactorSetCUR = PartialFactorSetCUR.CLASSI
    stabilityec7separtialfactorset: int = 0  # fixed value
    overallstabilitydrained: int = 1  # fixed value

    # These are all subgroups (key=value)
    partial_factors_eurocode_da1_set1: PartialFactorsEurocodeDa1Set1 = (
        PartialFactorsEurocodeDa1Set1()
    )
    partial_factors_eurocode_da1_set2: PartialFactorsEurocodeDa1Set2 = (
        PartialFactorsEurocodeDa1Set2()
    )
    partial_factors_eurocode_da2: PartialFactorsEurocodeDa2 = PartialFactorsEurocodeDa2()
    partial_factors_eurocode_da3: PartialFactorsEurocodeDa3 = PartialFactorsEurocodeDa3()
    partial_factors_ec7_nl_0: PartialFactorsEc7Nl0 = PartialFactorsEc7Nl0()
    partial_factors_ec7_nl_1: PartialFactorsEc7Nl1 = PartialFactorsEc7Nl1()
    partial_factors_ec7_nl_2: PartialFactorsEc7Nl2 = PartialFactorsEc7Nl2()
    partial_factors_ec7_nl_3: PartialFactorsEc7Nl3 = PartialFactorsEc7Nl3()
    partial_factors_ec7_b_set1: PartialFactorsEc7BSet1 = PartialFactorsEc7BSet1()
    partial_factors_ec7_b_set2: PartialFactorsEc7BSet2 = PartialFactorsEc7BSet2()
    partial_factors_cur_i: PartialFactorsCurI = PartialFactorsCurI()
    partial_factors_cur_ii: PartialFactorsCurIi = PartialFactorsCurIi()
    partial_factors_cur_iii: PartialFactorsCurIii = PartialFactorsCurIii()
    partial_factors_ec7_se_vk1: PartialFactorsEc7SeVk1 = PartialFactorsEc7SeVk1()
    partial_factors_ec7_se_vk2: PartialFactorsEc7SeVk2 = PartialFactorsEc7SeVk2()
    partial_factors_ec7_se_vk3: PartialFactorsEc7SeVk3 = PartialFactorsEc7SeVk3()


class SheetPileElement(DSeriesUnmappedNameProperties):
    name: constr(min_length=1, max_length=50) = _DEFAULT_SHEET_PILING_ELEMENT_NAME
    sheetpilingelementmaterialtype: SheetPilingElementMaterialType = (
        SheetPilingElementMaterialType.Steel
    )
    sheetpilingelementei: confloat(ge=0.001, le=1e12) = 100000
    sheetpilingelementwidth: confloat(ge=0, le=1000) = 1
    sheetpilingelementlevel: confloat(ge=-10000, le=10000) = -10
    sheetpilingelementheight: conint(ge=10, le=100000) = 400
    sheetpilingelementcoatingarea: confloat(ge=0.01, le=10) = 1.35
    sheetpilingpilewidth: confloat(ge=0, le=100000) = 0
    sheetpilingelementsectionarea: conint(ge=10, le=100000) = 170
    sheetpilingelementresistingmoment: conint(ge=0, le=100000) = 0
    sheetpilingelementreductionfactorei: confloat(ge=0.01, le=10) = 1
    sheetpilingelementnote: constr(min_length=0, max_length=20) = ""
    sheetpilingelementmaxcharacteristicmoment: confloat(ge=0, le=100000) = 0
    sheetpilingelementmaxplasticcharacteristicmoment: confloat(ge=0, le=100000) = 0
    sheetpilingelementkmod: confloat(ge=0.01, le=1) = 0.01
    sheetpilingelementmaterialfactor: confloat(ge=0.01, le=10) = 0.01
    ssheetpilingelementreductionfactormaxmoment: confloat(ge=0.01, le=10) = 1
    diaphragmwallissymmetric: conint(ge=0, le=1) = 0
    diaphragmwallposeielastoplastic1: confloat(ge=0, le=100000) = 0
    diaphragmwallnegeielastoplastic1: confloat(ge=0, le=100000) = 0
    diaphragmwallposmomelastic: confloat(ge=0, le=1000000000000) = 100000
    diaphragmwallnegmomelastic: confloat(ge=0, le=100000) = 0
    diaphragmwallposmomplastic: confloat(ge=0, le=100000) = 0
    diaphragmwallnegmomplastic: confloat(ge=0, le=100000) = 0
    diaphragmwallposeielastoplastic2: confloat(ge=0, le=100000) = 0
    diaphragmwallposmomelastoplastic: confloat(ge=0, le=100000) = 0
    diaphragmwallnegeielastoplastic2: confloat(ge=0, le=100000) = 0
    diaphragmwallnegmomelastoplastic: confloat(ge=0, le=100000) = 0
    woodensheetpilingelemente: confloat(ge=0.001, le=1000000000000) = 100000
    woodensheetpilingelementcharacflexuralstrength: confloat(ge=0, le=100000) = 0
    woodensheetpilingelementksys: confloat(ge=0.01, le=10) = 1.15
    woodensheetpilingelementkdef: confloat(ge=0.01, le=10) = 1
    woodensheetpilingelementpsi2eff: confloat(ge=0, le=10) = 1
    woodensheetpilingelementmaterialfactor: confloat(ge=0.01, le=10) = 1.3
    woodensheetpilingelementkmodfshort: confloat(ge=0.01, le=10) = 0.65
    woodensheetpilingelementkmodflong: confloat(ge=0.01, le=10) = 0.5
    woodensheetpilingelementkmode: confloat(ge=0.01, le=10) = 0.8

    def __init__(self, *args, **kwargs) -> "SheetPileElement":
        default_dict = dict(self.__fields__)
        for field, value in kwargs.items():
            if value is None:
                kwargs[field] = default_dict[field].default
        super().__init__(*args, **kwargs)

    def update_arguments_if_invalid_input(self) -> None:
        """This function updates sheetpilingelementkmod and
        result_dict_gamma_m_value according to the
        sheetpilingelementmaterialtype inputed by the user.

        For some types the user can define their own parameters and for
        some others default values are defined.
        """
        # k-mod value
        result_dict_k_mod_value = {
            0: self.sheetpilingelementkmod,
            1: 1.0,
            2: 1.0,
            3: 1.0,
            4: self.sheetpilingelementkmod,
            5: self.sheetpilingelementkmod,
        }
        if (
            self.sheetpilingelementkmod
            is not result_dict_k_mod_value[self.sheetpilingelementmaterialtype.value]
        ):
            self.sheetpilingelementkmod = result_dict_k_mod_value[
                self.sheetpilingelementmaterialtype.value
            ]
            warnings.warn(
                "Inputed value for kmod was not valid and was change to "
                + str(self.sheetpilingelementkmod),
                ImportWarning,
            )
        # gammma_m value
        result_dict_gamma_m_value = {
            0: self.sheetpilingelementmaterialfactor,
            1: 1.0,
            2: 1.1,
            3: 1.0,
            4: 1.2,
            5: self.sheetpilingelementmaterialfactor,
        }
        if (
            self.sheetpilingelementmaterialfactor
            is not result_dict_gamma_m_value[self.sheetpilingelementmaterialtype.value]
        ):
            self.sheetpilingelementmaterialfactor = result_dict_gamma_m_value[
                self.sheetpilingelementmaterialtype.value
            ]
            warnings.warn(
                "Inputed value for material factor was not valid and was change to "
                + str(self.sheetpilingelementmaterialfactor),
                ImportWarning,
            )


class SheetPiling(DSeriesStructureCollection):
    sheetpiling: conlist(SheetPileElement, min_items=1) = [SheetPileElement()]
    leveltopsheetpiling: confloat(ge=-10000, le=10000) = 0.0
    lengthsheetpiling: confloat(gt=0) = 10


class Anchor(DSheetpilingTableEntry):
    name: constr(min_length=1, max_length=50)
    level: float = 0
    e_modulus: confloat(gt=0) = 2.1e8
    cross_section: Optional[confloat(gt=0)] = 1e-3
    wall_height_kranz: Optional[confloat(ge=0)] = 0.00
    length: Optional[confloat(gt=0)] = 1
    angle: Optional[float] = 0.00
    yield_force: Optional[confloat(ge=0)] = 0.00
    side: Side = Side.RIGHT


class Anchors(DSheetpilingUnwrappedTable):
    """Container for Anchor."""

    anchors: List[Anchor] = []

    @property
    def anchor_names(self) -> Set[str]:
        return {anchor.name for anchor in self.anchors}


class AnchorOrStrutPresstressReference(BaseDataClass):
    """Used in the ConstructionStage to identify anchor and assign a
    prestress."""

    name: str
    pre_stress: confloat(ge=0) = _DEFAULT_PRE_STRESS


class Strut(DSheetpilingTableEntry):
    name: constr(min_length=1, max_length=50)
    level: float = 0
    e_modulus: confloat(gt=0) = 2.1e8
    cross_section: Optional[confloat(gt=0)] = 1e-4
    length: Optional[confloat(gt=0)] = 1
    angle: Optional[float] = 0.00
    buckling_force: Optional[confloat(ge=0)] = 0.00
    side: Side = Side.RIGHT


class Struts(DSheetpilingUnwrappedTable):
    """Container for Strut."""

    struts: List[Strut] = []

    @property
    def strut_names(self) -> Set[str]:
        return {strut.name for strut in self.struts}


class ConstructionStage(DSeriesUnmappedNameProperties):
    name: constr(min_length=1, max_length=50)
    displacement_top_sheet_piling: float = 0
    passive_side: PassiveSide = PassiveSide.DSHEETPILING_DETERMINED
    method_left: LateralEarthPressureMethodStage = (
        LateralEarthPressureMethodStage.KA_KO_KP
    )
    method_right: LateralEarthPressureMethodStage = (
        LateralEarthPressureMethodStage.KA_KO_KP
    )
    water_level_left: str = _DEFAULT_WATER_LEVEL_NAME
    water_level_right: str = _DEFAULT_WATER_LEVEL_NAME
    surface_left: str = _DEFAULT_SURFACE_NAME
    surface_right: str = _DEFAULT_SURFACE_NAME
    soil_profile_left: Optional[str] = _DEFAULT_SOIL_PROFILE_NAME
    soil_profile_right: Optional[str] = _DEFAULT_SOIL_PROFILE_NAME
    anchors: List[AnchorOrStrutPresstressReference] = []
    struts: List[AnchorOrStrutPresstressReference] = []
    spring_supports: List[str] = []
    rigid_supports: List[str] = []
    uniform_loads: List[str] = []
    surcharge_loads_left: List[str] = []
    surcharge_loads_right: List[str] = []
    horizontal_line_loads: List[str] = []
    moment_loads: List[str] = []
    normal_forces: List[str] = []


class ConstructionStages(DSeriesStructureCollection):
    stages: List[ConstructionStage] = []

    @property
    def stage_names(self) -> Set[str]:
        return {stage.name for stage in self.stages}


class WaterLevel(DSeriesNoParseSubStructure):
    name: constr(min_length=1, max_length=50)
    level: float
    distribution_type: DistributionType = DistributionType.NONE
    standard_deviation: float = 0.0


class WaterLevels(DSeriesNoParseSubStructure):
    levels: List[WaterLevel] = []

    @property
    def water_level_names(self) -> Set[str]:
        return {water_level.name for water_level in self.levels}


class StageOptions(DSeriesInlineMappedProperties):
    """Representation of [STAGE] block."""

    stagepartialfactorsetcur: PartialFactorSetCUR = PartialFactorSetCUR.UNKNOWN
    stageverify: int = 0
    stageanchorfactor: confloat(ge=0.001, le=1000) = 1
    stagepartialfactorsetec7nadnl: PartialFactorSetEC7NADNL = PartialFactorSetEC7NADNL.RC0
    stageverifyec7nadnl: int = 0
    stageanchorfactorec7nadnl: confloat(ge=0.001, le=1000) = 1
    stageverifyec7nadb: int = 0
    stagepartialfactorsetec7nadse: int = 0  # fixed value
    stageverifyec7nadse: int = 0  # fixed value


class CalculationOptionsPerStage(DSeriesStructureCollection):
    """Representation of [CALCULATION OPTIONS PER STAGE] block."""

    stageoptions: List[StageOptions] = []


class UniformLoad(DSeriesUnmappedNameProperties):
    name: constr(min_length=1, max_length=50)
    uniformloadleft: float
    uniformloadright: float
    uniformloadpermanent: LoadTypePermanentVariable = LoadTypePermanentVariable.PERMANENT
    uniformloadfavourable: LoadTypeFavourableUnfavourable = (
        LoadTypeFavourableUnfavourable.DSHEETPILING_DETERMINED
    )
    uniformloadstandarddeviationleft: float
    uniformloadstandarddeviationright: float
    uniformloaddistleft: DistributionType = DistributionType.NORMAL
    uniformloaddistright: DistributionType = DistributionType.NORMAL


class UniformLoads(DSeriesStructureCollection):
    loads: List[UniformLoad] = []

    @property
    def load_names(self) -> Set[str]:
        return {load.name for load in self.loads}


class SurchargePoint(DSeriesInlineMappedProperties):
    surchargeloaddistance: float
    surchargeloadvalue: float


class SurchargeLoad(DSheetpilingSurchargeLoad):
    name: constr(min_length=1, max_length=50)
    points: List[SurchargePoint] = []
    surchargeloadpermanent: LoadTypePermanentVariable = (
        LoadTypePermanentVariable.PERMANENT
    )
    surchargeloadfavourable: LoadTypeFavourableUnfavourable = (
        LoadTypeFavourableUnfavourable.DSHEETPILING_DETERMINED
    )
    surchargeloadstandarddeviation: float = 0.00
    surchargeloaddistribution: DistributionType = DistributionType.NORMAL


class SurchargeLoads(DSeriesStructureCollection):
    loads: List[SurchargeLoad] = []

    @property
    def load_names(self) -> Set[str]:
        return {load.name for load in self.loads}


class Surface(DSeriesNoParseSubStructure):  # TODO determine structure
    name: constr(min_length=1, max_length=50)
    points: conlist(Point, min_items=1)
    points: conlist(dict, min_items=1)
    distribution_type: DistributionType = DistributionType.NONE
    std: confloat(ge=0.0) = 0.0


class Surfaces(DSeriesNoParseSubStructure):  # TODO GroupList should be suitable?
    surfaces: List[Surface] = []

    @property
    def surface_names(self) -> Set[str]:
        return {surface.name for surface in self.surfaces}


class HorizontalLineLoad(DSeriesNoParseSubStructure):
    name: constr(min_length=1, max_length=50)
    level: confloat(ge=-1e12, le=1e12) = 0
    load: confloat(ge=-1e12, le=1e12) = 100
    load_type: LoadTypeFavourableUnfavourable = (
        LoadTypeFavourableUnfavourable.DSHEETPILING_DETERMINED
    )
    duration_type: LoadTypePermanentVariable = LoadTypePermanentVariable.PERMANENT


class HorizontalLineLoads(DSeriesNoParseSubStructure):
    loads: List[HorizontalLineLoad] = []


class Moment(DSeriesNoParseSubStructure):
    name: constr(min_length=1, max_length=50)
    level: confloat(ge=-1e12, le=1e12) = 0
    load: confloat(ge=-1e12, le=1e12) = 100
    load_type: LoadTypeFavourableUnfavourableMoment = (
        LoadTypeFavourableUnfavourableMoment.FAVOURABLE
    )
    duration_type: LoadTypePermanentVariable = LoadTypePermanentVariable.PERMANENT


class Moments(DSeriesNoParseSubStructure):
    loads: List[Moment] = []


class NormalForce(DSeriesNoParseSubStructure):
    name: constr(min_length=1, max_length=50)
    force_at_sheet_pile_top: confloat(ge=-1e12, le=1e12)
    force_at_surface_level_left_side: confloat(ge=-1e12, le=1e12)
    force_at_surface_level_right_side: confloat(ge=-1e12, le=1e12)
    force_at_sheet_pile_toe: confloat(ge=-1e12, le=1e12)
    load_type: LoadTypeFavourableUnfavourableMoment = (
        LoadTypeFavourableUnfavourableMoment.FAVOURABLE
    )
    duration_type: LoadTypePermanentVariable = LoadTypePermanentVariable.PERMANENT


class NormalForces(DSeriesNoParseSubStructure):
    loads: List[NormalForce] = []


class Support(DSeriesNoParseSubStructure):
    """Internal structure for spring and rigid supports."""

    name: constr(min_length=1, max_length=50)
    level: float
    rotational_stiffness: confloat(ge=0)
    translational_stiffness: confloat(ge=0)


class SupportContainer(DSeriesNoParseSubStructure):
    supports: List[Support] = []

    @property
    def support_names(self) -> Set[str]:
        return {support.name for support in self.supports}


class Water(DSeriesUnmappedNameProperties):
    unit_weight_of_water: float = 9.81


class DSheetPilingInputStructure(DSeriesStructure):
    """Representation of complete .shi file."""

    soil_collection: SoilCollection = SoilCollection()
    run_identification: str = 2 * "\n"
    model: Model = Model()
    cpt_list: str = cleandoc(
        """
        Count=0
        """
    )
    sheet_piling: Union[str, SheetPiling] = SheetPiling()
    combined_wall: str = ""
    vertical_balance: str = cleandoc(
        """
        SheetPilingQcRep=0.000
        SheetPilingXi=1.39
        """
    )
    settlement_by_vibration_params: str = cleandoc(
        """
        SheetPilingNumberOfPilesDrilled=2
        """
    )
    horizontal_line_loads: Union[HorizontalLineLoads, str, None] = None
    uniform_loads: Optional[UniformLoads] = None
    surcharge_loads: Optional[SurchargeLoads] = None
    water: str = ""
    earth_quake: str = cleandoc(
        """
        0.00
        """
    )
    soil_profiles: Union[SoilProfiles, str] = cleandoc(
        f"""
          1 Number of spring characteristics curves
          0 1/0 : Yes/No Unloading curve
          1 1/0 : Yes/No Use Cur values

          1 Number of soil profiles 
        {_DEFAULT_SOIL_PROFILE_NAME}
               0.000 X coordinate
               0.000 Y coordinate

          1 Number of soil layers per soil profile
         Nr      Level   Wosp Top  Wosp Bott Material Name
          1       0.00       0.00       0.00 New Material

        """
    )
    surfaces: Union[str, Surfaces] = cleandoc(
        f"""
        1 Number of surfaces 
        1     1 {_DEFAULT_SURFACE_NAME}
        0.00  Standard deviation
        2  Distribution type
        Nr     X-coord     Value
        1       0.00       0.00
        """
    )
    water: Water = Water()
    waterlevels: Union[WaterLevels, str] = cleandoc(
        f"""
          1 Number of Waterlevels 
          3 Number of Data per Waterlevel 
        {_DEFAULT_WATER_LEVEL_NAME}
              0.00
              0.00
                 2

        """
    )
    construction_stages: Union[str, ConstructionStages] = ConstructionStages()
    calculation_options_per_stage: CalculationOptionsPerStage = (
        CalculationOptionsPerStage()
    )
    calculation_options: CalculationOptions = CalculationOptions()
    probabilistic_calculation_options: str = cleandoc(
        """
        ZFunctionType=1
        ZFunctionMobilisation=100.00
        ZFunctionMoment=0.00
        ZFunctionAnchorForce=0.00
        ZFunctionAnchor=-1
        """
    )
    probabilistic_control_parameters: str = cleandoc(
        """
        IterationFactor=1.000
        IterationNumber=15
        CriteriaLimit1=0.010000
        CriteriaLimit2=0.010000
        ReliabilityStage=0
        """
    )

    """
    When there are no anchors, struts, supports, or soil displacements in the model,
    their respective blocks is not present in the .shi file.
    """
    soil_displacements: Optional[str] = None
    rigid_supports: Union[str, SupportContainer, None] = None
    spring_supports: Union[str, SupportContainer, None] = None
    moments: Union[str, Moments, None] = None
    normal_forces: Union[str, NormalForces, None] = None
    anchors: Union[str, Anchors, None] = None
    struts: Union[str, Struts, None] = None

    # Custom validator
    _validate_run_identification = make_newline_validator(
        "run_identification", req_newlines=REQ_RUN_LINES
    )

    @property
    def is_valid(self) -> bool:
        return self._validator().is_valid

    def has_stage(self, stage_id: int) -> bool:
        try:
            self.construction_stages.stages[stage_id]
            return True
        except IndexError:
            return False

    def _validator(self) -> DSheetPilingValidator:
        return DSheetPilingValidator(self)

    def set_model(self, **kwargs) -> None:
        kwargs = self._filter_none_values_from_key_value_dict(kwargs)
        self.model = Model(**kwargs)

        # TODO Singlepile model is a won't have, raise error here?

        if self.model.method != LateralEarthPressureMethod.MIXED:
            for stage in self.construction_stages.stages:
                stage.method_left = (
                    LateralEarthPressureMethodStage.get_stage_type_from_method(
                        self.model.method
                    )
                )
                stage.method_right = (
                    LateralEarthPressureMethodStage.get_stage_type_from_method(
                        self.model.method
                    )
                )

    def set_calculation_options(self, **kwargs) -> None:
        kwargs = self._filter_none_values_from_key_value_dict(kwargs)
        # here add function that turns kwargs from snake case to pascal case
        kwargs = self._from_snake_to_pascal_case(kwargs)
        self.calculation_options = CalculationOptions(**kwargs)

    def add_calculation_options_per_stage(
        self,
        input_calc_options: CalculationOptionsPerStageExternal,
        stage_id: int,
    ) -> None:
        _map_external_to_internal_values = {
            VerifyType.CUR: {
                "stagepartialfactorsetcur": input_calc_options.partial_factor_set,
                "stageverify": stage_id + 1,
                "stageanchorfactor": input_calc_options.anchor_factor,
            },
            VerifyType.EC7NL: {
                "stagepartialfactorsetec7nadnl": input_calc_options.partial_factor_set,
                "stageverifyec7nadnl": stage_id + 1,
                "stageanchorfactorec7nadnl": input_calc_options.anchor_factor,
            },
            VerifyType.EC7BE: {"stageverifyec7nadb": stage_id + 1},
        }
        stageoptions = StageOptions(
            **_map_external_to_internal_values[self.calculation_options.verifytype]
        )
        self.calculation_options_per_stage.stageoptions[stage_id] = stageoptions

    def set_curve_settings(self, curve_settings: CurveSettings) -> None:
        if isinstance(self.soil_profiles, str):
            self.soil_profiles = SoilProfiles()
        self.soil_profiles.curve_number = curve_settings.curve_number
        self.soil_profiles.use_unloading_reloading_curve = (
            curve_settings.use_unloading_reloading_curve
        )
        self.soil_profiles.modulus_reaction_type = (
            curve_settings.modulus_reaction_type.value
        )

    def add_stage(
        self,
        name: str,
        passive_side: PassiveSide,
        method_left: LateralEarthPressureMethodStage,
        method_right: LateralEarthPressureMethodStage,
        pile_top_displacement: Optional[float],
    ) -> None:
        if isinstance(self.construction_stages, str):
            self.construction_stages = ConstructionStages()
        if name in self.construction_stages.stage_names:
            raise ValueError(
                f"Stage name {name} already present: all stage names must be unique"
            )

        method_left = self._get_validated_lateral_earth_pressure_method(method_left)
        method_right = self._get_validated_lateral_earth_pressure_method(method_right)
        self._validate_lateral_earth_pressure_methods_for_single_pile_model(
            method_left, method_right
        )

        kwargs = self._filter_none_values_from_key_value_dict(
            {
                "name": name,
                "displacement_top_sheet_piling": pile_top_displacement,
                "passive_side": passive_side,
                "method_left": method_left,
                "method_right": method_right,
            }
        )
        self.construction_stages.stages.append(ConstructionStage(**kwargs))
        # defaults for calculation options should be initialised here
        self.calculation_options_per_stage.stageoptions.append(StageOptions())

    def _get_validated_lateral_earth_pressure_method(
        self, stage_method: Optional[LateralEarthPressureMethodStage]
    ) -> LateralEarthPressureMethodStage:
        """When no stage_method is provided, returns the method set in the
        model options.

        Raises ValueError if method is not compatible with method set on
        model.
        """
        if not isinstance(self.model, Model):
            raise ValueError(f"Model is a string; set model first")
        if stage_method is None:
            return self.model.method
        error = ValueError(
            f"stage_method {stage_method} is not compatible with method on model {self.model.method}"
        )
        if (
            self.model.method.name == LateralEarthPressureMethodStage.KA_KO_KP.name
            and stage_method.name == LateralEarthPressureMethodStage.C_PHI_DELTA.name
        ):
            raise error
        if (
            self.model.method.name == LateralEarthPressureMethodStage.C_PHI_DELTA.name
            and stage_method.name == LateralEarthPressureMethodStage.KA_KO_KP.name
        ):
            raise error
        return stage_method

    def _validate_lateral_earth_pressure_methods_for_single_pile_model(
        self,
        method_left: LateralEarthPressureMethodStage,
        method_right: LateralEarthPressureMethodStage,
    ) -> None:
        if not isinstance(self.model, Model):
            raise ValueError(f"Model is a string; set model first")
        if self.model.model == ModelType.SINGLE_PILE:
            if method_left != method_right:
                raise ValueError(
                    f"For the Single Pile Model the left and right method must be the same. Received: left {method_left} and right {method_right}"
                )

    def _filter_none_values_from_key_value_dict(self, dict_: Dict) -> Dict:
        # This way defaults can be defined only in the internal.
        return {k: v for k, v in dict_.items() if v is not None}

    def _from_snake_to_pascal_case(self, dict_: Dict) -> Dict:
        return {k.replace("_", ""): v for k, v in dict_.items()}

    def add_water_level(self, stage_id: int, water_level: WaterLevel, side: Side) -> None:
        # Add water level to water levels
        if isinstance(self.waterlevels, str):
            self.waterlevels = WaterLevels()
        if water_level.name not in self.waterlevels.water_level_names:
            self.waterlevels.levels.append(water_level)

        # Add water level to stages.
        stage = self.construction_stages.stages[stage_id]
        if side == Side.BOTH:
            stage.water_level_left = stage.water_level_right = water_level.name
        elif side == Side.LEFT:
            stage.water_level_left = water_level.name
        elif side == Side.RIGHT:
            stage.water_level_right = water_level.name
        else:
            raise ValueError(f"Provide a Side, received {side}")

    def add_profile(self, stage_id: int, profile: SoilProfile, side: Side) -> None:
        # Add profile to profiles
        if isinstance(self.soil_profiles, str):
            self.soil_profiles = SoilProfiles()
        if profile.name not in self.soil_profiles.soil_profile_names:
            self.soil_profiles.soil_profiles.append(profile)

        soil_names = self.soil_collection.soil_names
        for i, layer in enumerate(profile.layers):
            if layer.soil not in soil_names:
                raise ValueError(
                    f"Soil {layer.soil} of layer {i} is not in soil collection; add soil first via model."
                )
        # Add profile to stages.
        if side == Side.BOTH:
            self.construction_stages.stages[stage_id].soil_profile_left = profile.name
            self.construction_stages.stages[stage_id].soil_profile_right = profile.name
        elif side == Side.LEFT:
            self.construction_stages.stages[stage_id].soil_profile_left = profile.name
        elif side == Side.RIGHT:
            self.construction_stages.stages[stage_id].soil_profile_right = profile.name
        else:
            raise ValueError(f"Provide a Side, received {side}")

    def add_surface(self, stage_id: int, surface: Surface, side: Side) -> None:
        # Add surface to surfaces
        if isinstance(self.surfaces, str):
            self.surfaces = Surfaces()
        if surface.name not in self.surfaces.surface_names:
            self.surfaces.surfaces.append(surface)

        # Add surface to stages.
        if side == Side.BOTH:
            self.construction_stages.stages[stage_id].surface_left = surface.name
            self.construction_stages.stages[stage_id].surface_right = surface.name
        elif side == Side.LEFT:
            self.construction_stages.stages[stage_id].surface_left = surface.name
        elif side == Side.RIGHT:
            self.construction_stages.stages[stage_id].surface_right = surface.name
        else:
            raise ValueError(f"Provide a Side, received {side}")

    def set_construction(
        self, top_level: float, elements: List[SheetPileElement]
    ) -> None:
        elements.sort(key=lambda element: element.sheetpilingelementlevel, reverse=True)
        for sheet in elements:
            sheet.update_arguments_if_invalid_input()
        try:
            bottom_level = elements[-1].sheetpilingelementlevel
        except IndexError:
            raise ValueError("At least one sheet should be provided")

        self.sheet_piling = SheetPiling(
            sheetpiling=elements,
            leveltopsheetpiling=top_level,
            lengthsheetpiling=top_level - bottom_level,
        )

    def add_element_in_sheet_piling(
        self,
        sheet: Any,
        location_top: Optional[Point] = None,
    ) -> None:
        self.sheet_piling.update_level_top_sheet_pile(location_top)
        try:
            sheet = sheet._to_internal()
        except AttributeError:
            raise ValueError(f"Invalid argument, expected sheet, got {type(sheet)}")
        sheet.update_arguments_if_invalid_input()
        self.sheet_piling.sheetpiling.append(sheet)
        self.sheet_piling.update_length_of_sheet_pile()

    def add_anchor(self, stage_id: int, anchor: Anchor, pre_tension: float) -> None:
        if not isinstance(self.anchors, Anchors):
            self.anchors = Anchors()
        if anchor.name not in self.anchors.anchor_names:
            self.anchors.anchors.append(anchor)

        self.construction_stages.stages[stage_id].anchors.append(
            AnchorOrStrutPresstressReference(
                name=anchor.name, pre_stress=pre_tension or _DEFAULT_PRE_STRESS
            )
        )

    def add_strut(self, stage_id: int, strut: Strut, pre_compression: float) -> None:
        if not isinstance(self.struts, Struts):
            self.struts = Struts()
        if strut.name not in self.struts.strut_names:
            self.struts.struts.append(strut)

        self.construction_stages.stages[stage_id].struts.append(
            AnchorOrStrutPresstressReference(
                name=strut.name, pre_stress=pre_compression or _DEFAULT_PRE_STRESS
            )
        )

    def add_load(
        self, load: Union[HorizontalLineLoad, Moment, NormalForce], stage_id: int
    ):
        if isinstance(load, HorizontalLineLoad):
            self.is_valid_unique_load_names(
                load_list=self.construction_stages.stages[stage_id].horizontal_line_loads,
                name=load.name,
            )
            if (
                isinstance(self.horizontal_line_loads, str)
                or self.horizontal_line_loads is None
            ):
                self.horizontal_line_loads = HorizontalLineLoads()
            if load not in self.horizontal_line_loads.loads:
                self.horizontal_line_loads.loads.append(load)
            self.construction_stages.stages[stage_id].horizontal_line_loads.append(
                load.name
            )
        elif isinstance(load, Moment):
            self.is_valid_unique_load_names(
                load_list=self.construction_stages.stages[stage_id].moment_loads,
                name=load.name,
            )
            if isinstance(self.moments, str) or self.moments is None:
                self.moments = Moments()
            if load not in self.moments.loads:
                self.moments.loads.append(load)
            self.construction_stages.stages[stage_id].moment_loads.append(load.name)
        elif isinstance(load, NormalForce):
            self.is_valid_unique_load_names(
                load_list=self.construction_stages.stages[stage_id].normal_forces,
                name=load.name,
            )
            if isinstance(self.normal_forces, str) or self.normal_forces is None:
                self.normal_forces = NormalForces()
            if load not in self.normal_forces.loads:
                self.normal_forces.loads.append(load)
            self.construction_stages.stages[stage_id].normal_forces.append(load.name)
        elif isinstance(load, UniformLoad):
            self.add_uniform_load(load=load, stage_id=stage_id)
        else:
            raise ValueError(f"This type of load is not yet supported")

    def add_uniform_load(self, stage_id: int, load: UniformLoad) -> None:
        self.is_valid_unique_load_names(
            load_list=self.construction_stages.stages[stage_id].uniform_loads,
            name=load.name,
        )
        if not isinstance(self.uniform_loads, UniformLoads):
            self.uniform_loads = UniformLoads()
        if load not in self.uniform_loads.loads:
            self.uniform_loads.loads.append(load)
        self.construction_stages.stages[stage_id].uniform_loads.append(load.name)

    def add_surcharge_load(self, stage_id: int, load: SurchargeLoad, side: Side) -> None:
        if not isinstance(self.surcharge_loads, SurchargeLoads):
            self.surcharge_loads = SurchargeLoads()
        if load not in self.surcharge_loads.loads:
            self.surcharge_loads.loads.append(load)
        # Add load to stages.
        stage = self.construction_stages.stages[stage_id]
        if side == Side.BOTH:
            self.is_valid_unique_load_names(
                load_list=stage.surcharge_loads_left, name=load.name
            )
            stage.surcharge_loads_left.append(load.name)
            self.is_valid_unique_load_names(
                load_list=stage.surcharge_loads_right, name=load.name
            )
            stage.surcharge_loads_right.append(load.name)
        elif side == Side.LEFT:
            self.is_valid_unique_load_names(
                load_list=stage.surcharge_loads_left, name=load.name
            )
            stage.surcharge_loads_left.append(load.name)
        elif side == Side.RIGHT:
            self.is_valid_unique_load_names(
                load_list=stage.surcharge_loads_right, name=load.name
            )
            stage.surcharge_loads_right.append(load.name)
        else:
            raise ValueError(f"Provide a Side, received {side}")

    def is_valid_unique_load_names(self, load_list: List[str], name: str) -> bool:
        """Load list should have unique names in list of loads."""
        if name in load_list:
            raise ValueError(
                f"{name} load name is duplicated. Please change the name of the load."
            )
        return True

    def add_spring_support(self, stage_id: int, support: Support) -> None:
        if not isinstance(self.spring_supports, SupportContainer):
            self.spring_supports = SupportContainer()
        if support.name not in self.spring_supports.support_names:
            self.spring_supports.supports.append(support)

        self.construction_stages.stages[stage_id].spring_supports.append(support.name)

    def add_rigid_support(self, stage_id: int, support: Support) -> None:
        if not isinstance(self.rigid_supports, SupportContainer):
            self.rigid_supports = SupportContainer()
        if support.name not in self.rigid_supports.support_names:
            self.rigid_supports.supports.append(support)

        self.construction_stages.stages[stage_id].rigid_supports.append(support.name)


class Resume(DSheetpilingWithNumberOfRowsTable):
    resume: List[Dict[str, float]]


class BreukData(DSeriesInlineReversedProperties):
    percentage_mobilized_resistance_left: float
    percentage_mobilized_resistance_right: float
    effective_left: float
    effective_right: float
    water_pressure_left: float
    water_pressure_right: float
    max_passive_effect_resistance_left: float
    max_passive_effect_resistance_right: float
    max_passive_moment_left: float
    max_passive_moment_right: float
    max_mobilized_passive_moment_left: float
    max_mobilized_passive_moment_right: float
    vertical_force_left: float
    vertical_force_right: float
    level_of_single_support: float
    node_of_single_support: float
    moment_present: float
    active_side: float


class MomentsForcesDisplacements(DSeriesWrappedTableStructure):
    momentsforcesdisplacements: List[Dict[str, float]]


class AnchorData(DSheetpilingWithNumberOfRowsTable):
    anchordata: List[Dict[str, Union[float, str]]]


class Pressures(DSheetpilingWithNumberOfRowsTable):
    pressures: List[Dict[str, Union[float, str]]]


class SideOutput(DSeriesStructure):
    calculation_method: Optional[str]
    water_level: Optional[str]
    surface: Optional[str]
    soil_profile_for_single_pile_model: Optional[str]
    soil_profile_for_sheet_piling_model: Optional[str]
    pressures: Optional[Pressures]
    force_from_layer: Optional[str]
    lambdas: Optional[str]
    slide_plane: Optional[str]
    vertical_balance_per_layer: Optional[str]


class OutputConstructionStage(DSeriesRepeatedGroupedProperties):
    anchor_data: Optional[AnchorData]
    hload_data: Optional[str]
    breuk_data: Optional[BreukData]
    moments_forces_displacements: Optional[MomentsForcesDisplacements]
    side: Optional[List[SideOutput]]
    uniform_load_data: Optional[str]
    horizontal_line_load_data: Optional[str]
    surcharge_data: Optional[str]
    normal_force_data: Optional[str]
    moment_data: Optional[str]
    support_data: Optional[str]
    vertical_balance_data: Optional[str]

    @classmethod
    def get_list_field_names(cls) -> List[str]:
        return ["side"]


class DesignLengthInfo(DSeriesWrappedTableStructure):
    designlengthinfo: List[Dict[str, float]]


class DesignLengthCalculation(DSeriesWrappedTableStructure):
    designlengthcalculation: List[Dict[str, float]]


class DesignSheetpileLength(DSeriesStructure):
    design_length_info: Optional[DesignLengthInfo]
    design_length_calculation: Optional[DesignLengthCalculation]


class PointsOnSheetpile(DSheetpilingWithNumberOfRowsTable):
    pointsonsheetpile: List[Dict[str, float]]


class CurAnchorForceResults(DSheetpilingWithNumberOfRowsTable):
    curanchorforceresults: List[Dict[str, float]]


class BaseVerificationStructureProperties(DSeriesRepeatedGroupedProperties):
    points_on_sheetpile: Optional[List[PointsOnSheetpile]]
    construction_stage: Optional[List[OutputConstructionStage]]
    resume: Optional[Resume]
    warning: Optional[str]
    error: Optional[str]

    # TODO Check how many optional variations exist
    basis_step_of_step_9____1_per_anchor: Optional[str]
    verify_anchor_force_step_9____1_based_on_step_6____5: Optional[str]
    verify_anchor_force_step_9____1_based_on_step_6____3: Optional[str]

    @classmethod
    def get_list_field_names(cls) -> List[str]:
        return ["points_on_sheetpile", "construction_stage"]


class VerifyAnchorForce(BaseVerificationStructureProperties):
    anchor_number: Optional[int]
    points_on_sheetpile: Optional[List[PointsOnSheetpile]]
    construction_stage: Optional[List[OutputConstructionStage]]
    resume: Optional[Resume]


class VerifySheetpileData(DSeriesStructure):
    verify_deformation: Optional[BaseVerificationStructureProperties]
    verify_moment_low_angle_of_subgr_reac: Optional[BaseVerificationStructureProperties]
    verify_moment_high_angle_of_subgr_reac: Optional[BaseVerificationStructureProperties]
    verify_low_mod_with_alt_passive_waterlevel: Optional[
        BaseVerificationStructureProperties
    ]
    verify_high_mod_with_alt_passive_waterlevel: Optional[
        BaseVerificationStructureProperties
    ]
    verify_anchor_force: Optional[VerifyAnchorForce]
    resume: Optional[Resume]
    factors_for_verification: Optional[str]


class DSheetPilingOutputStructure(DSeriesRepeatedGroupedProperties):
    @classmethod
    def get_list_field_names(cls) -> List[str]:
        return ["verify_sheetpile_data", "points_on_sheetpile", "construction_stage"]

    @classmethod
    def parse_text(cls, text):
        return super().parse_text(text)

    calculation_type: str

    # Sheetpile calculation
    sheet_pile_elements: Optional[str]
    design_sheetpile_length: Optional[DesignSheetpileLength]
    verify_sheetpile_data: Optional[List[VerifySheetpileData]]
    points_on_sheetpile: Optional[List[PointsOnSheetpile]]
    construction_stage: Optional[List[OutputConstructionStage]]

    # Vibration calculation
    settlement_by_vibration: Optional[str]

    # Verification calculation
    # TODO Split further based on A/B grouping
    overall_partial_factor_set: Optional[str]
    factors_for_overall_stability: Optional[str]
    overall_stability_results: Optional[str]

    calculation_overview: Optional[str]
    number_of_verifications: Optional[str]
    maximum_moment_effect_included: Optional[str]

    eurocode2: Optional[str]
    eurocode3: Optional[str]
    eurocode1_set1: Optional[str]
    eurocode1_set2: Optional[str]
    eurocode_belgie_set2: Optional[str]
    eurocode_belgie_set1: Optional[str]
    cur_anchor_force_results: Optional[CurAnchorForceResults]
    warning_list: Optional[str]
    warning: Optional[str]
    error: Optional[str]

    calculated_displacements: Optional[str]
    angles_kranz_calculation: Optional[str]
    kranz_calculation: Optional[str]

    # Verifications, duplicated for all verifications.
    verify_deformation: Optional[BaseVerificationStructureProperties]
    verify_moment_low_angle_of_subgr_reac: Optional[BaseVerificationStructureProperties]
    verify_moment_high_angle_of_subgr_reac: Optional[BaseVerificationStructureProperties]
    verify_low_mod_with_alt_passive_waterlevel: Optional[
        BaseVerificationStructureProperties
    ]
    verify_high_mod_with_alt_passive_waterlevel: Optional[
        BaseVerificationStructureProperties
    ]
    verify_anchor_force: Optional[BaseVerificationStructureProperties]
    resume: Optional[Resume]
    factors_for_verification: Optional[str]


class DSheetPilingStructure(DSeriesPilingParserStructure):
    version: Version = Version()
    version_externals: VersionExternals = VersionExternals()
    input_data: DSheetPilingInputStructure = DSheetPilingInputStructure()
    output_data: Optional[DSheetPilingOutputStructure] = None

    @property
    def is_valid(self) -> bool:
        return self.input_data.is_valid


class DSheetPilingDumpStructure(DSeriesPilingDumpParserStructure):
    dumpfile: DSheetPilingStructure



import abc
import re
from enum import Enum
from pathlib import Path
from typing import Dict, List, Optional, Set, Type, Union

from pydantic import DirectoryPath, FilePath

from geolib.geometry import Point
from geolib.models import BaseDataClass, BaseModel
from geolib.soils import Soil

from ...utils import camel_to_snake, snake_to_camel
from .analysis import DStabilityAnalysisMethod
from .dstability_parserprovider import DStabilityParserProvider
from .internal import (
    AnalysisType,
    BishopSlipCircleResult,
    CalculationType,
    DStabilityResult,
    DStabilityStructure,
    PersistablePoint,
    SoilCollection,
    SpencerSlipPlaneResult,
    Stage,
    UpliftVanSlipCircleResult,
    Waternet,
)
from .loads import Consolidation, DStabilityLoad
from .reinforcements import DStabilityReinforcement
from .serializer import DStabilityInputSerializer, DStabilityInputZipSerializer
from .states import DStabilityStateLinePoint, DStabilityStatePoint


class DStabilityCalculationType(Enum):
    """Set Type of Calculation."""

    BoundarySearch = 1
    SingleCalc = 2


class DStabilityCalculationModel(Enum):
    """Set Type of Calculation."""

    Bishop = 1
    UpliftVan = 2
    Spencer = 3


class DStabilityObject(BaseModel, metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def _to_dstability_sub_structure(self):
        raise NotImplementedError


class DStabilityModel(BaseModel):
    """D-Stability is software for soft soil slope stability.

    This model can read, modify and create
    .stix files
    """

    current_stage: int = -1
    datastructure: DStabilityStructure = DStabilityStructure()
    current_id: int = -1

    def __init__(self, *args, **data) -> None:
        super().__init__(*args, **data)
        self.current_id = self.datastructure.get_unique_id()

    @property
    def parser_provider_type(self) -> Type[DStabilityParserProvider]:
        return DStabilityParserProvider

    @property
    def console_path(self) -> Path:
        return Path("DStabilityConsole/D-Stability Console.exe")

    @property
    def soils(self) -> SoilCollection:
        """Enables easy access to the soil in the internal dict-like datastructure. Also enables edit/delete for individual soils."""
        return self.datastructure.soils

    def _get_next_id(self) -> int:
        self.current_id += 1
        return self.current_id

    def parse(self, *args, **kwargs):
        super().parse(*args, **kwargs)
        self.current_id = self.datastructure.get_unique_id()

    @property
    def waternets(self) -> List[Waternet]:
        return self.datastructure.waternets

    @property
    def output(self) -> DStabilityResult:
        # TODO Make something that works for all stages
        return self.get_result(self.current_stage)

    def get_result(self, stage_id: int) -> Dict:
        """
        Returns the results of a stage. Calculation results are based on analysis type and calculation type.

        Args:
            stage_id (int): Id of a stage.

        Returns:
            dict: Dictionary containing the analysis results of the stage.

        Raises:
            ValueError: No results or calculationsettings available
        """
        result = self._get_result_substructure(stage_id)
        return result  # TODO snake_case keys?

    def _get_result_substructure(self, stage_id: int) -> DStabilityResult:
        if self.datastructure.has_result(stage_id):
            result_id = self.datastructure.stages[stage_id].ResultId
            calculation_settings = self.datastructure.calculationsettings[stage_id]
            analysis_type = calculation_settings.AnalysisType
            calculation_type = calculation_settings.CalculationType

            results = self.datastructure.get_result_substructure(
                analysis_type, calculation_type
            )

            for result in results:
                if result.Id == result_id:
                    return result

        raise ValueError(f"No result found for result id {stage_id}")

    def get_slipcircle_result(
        self, stage_id: int
    ) -> Union[BishopSlipCircleResult, UpliftVanSlipCircleResult]:
        """
        Get the slipcircle(s) of the calculation result of a given stage.

        Args:
            stage_id (int): stage for which to get the available results

        Returns:
            Dict: dictionary of the available slipcircles per model for the given stage

        Raises:
            ValueError: Result is not available for provided stage id
            AttributeError: When the result has no slipcircle. Try get the slipplane
        """
        result = self._get_result_substructure(stage_id)
        return result.get_slipcircle_output()

    def get_slipplane_result(self, stage_id: int = 0) -> SpencerSlipPlaneResult:
        """
        Get the slipplanes of the calculations result of a stage.

        Args:
            stage_id (int): stage for which to get the available results

        Returns:
            dict: dictionary of the available slip planes per model for the given stage

        Raises:
            ValueError: Result is not available for provided stage id
            AttributeError: When the result has no slipplane. Try get the slipcircle
        """
        result = self._get_result_substructure(stage_id)
        return result.get_slipplane_output()

    def serialize(self, location: Union[FilePath, DirectoryPath]):
        """Support serializing to directory while developing for debugging purposes."""
        if not location.is_dir():
            serializer = DStabilityInputZipSerializer(ds=self.datastructure)
        else:
            serializer = DStabilityInputSerializer(ds=self.datastructure)
        serializer.write(location)
        self.filename = location

    def add_stage(self, label: str, notes: str, set_current=True) -> int:
        """Add a new stage to the model.

        Args:
            label: Label for the stage
            notes: Notes for the stage
            set_current: Whether to make the new stage the current stage.

        Returns:
            the id of the new stage
        """
        new_id = self._get_next_id()
        new_stage_id, new_unique_id = self.datastructure.add_default_stage(
            label, notes, new_id
        )

        if set_current:
            self.current_stage = new_stage_id
        self.current_id = new_unique_id
        return new_stage_id

    def copy_stage(self, label: str, notes: str, set_current=True) -> int:
        """Copy an existing stage and add it to the model.

        Args:
            label: Label for the stage
            notes: Notes for the stage
            set_current: Whether to make the new stage the current stage.

        Returns:
            the id of the new stage
        """
        new_id = self._get_next_id()
        new_stage_id, new_unique_id = self.datastructure.duplicate_stage(
            self.current_stage, label, notes, new_id
        )

        if set_current:
            self.current_stage = new_stage_id
        self.current_id = new_unique_id
        return new_stage_id

    @property
    def stages(self) -> List[Stage]:
        return self.datastructure.stages

    def add_point(self, point: Point, stage=None) -> int:
        """Add point, which should be unique in the model and return the created point id."""

    def add_soil(self, soil: Soil) -> int:
        """
        Add a new soil to the model. The code must be unique, the id will be generated

        Args:
            soil (Soil): a new soil

        Returns:
            int: id of the added soil
        """
        if self.soils.has_soilcode(soil.code):
            raise ValueError(f"The soil with code {soil.code} is already defined.")

        soil.id = self._get_next_id()
        persistant_soil = self.soils.add_soil(soil)
        return persistant_soil.Code

    def edit_soil(self, code: str, **kwargs: dict) -> None:
        """
        Edit an existing soil with parameter names based on the soil class members

        Args:
            code (str): the code of the soil
            kwargs (dict): the parameters and new values for example 'cohesion=2.0, friction_angel=25.0'

        Returns:
            bool: True for succes, False otherwise
        """
        return self.soils.edit_soil(code, **kwargs)

    @property
    def points(self):
        """Enables easy access to the points in the internal dict-like datastructure. Also enables edit/delete for individual points."""

    def add_layer(
        self,
        points: List[Point],
        soil_code: str,
        label: str = "",
        notes: str = "",
        stage_id: int = None,
    ) -> int:
        """
        Add a soil layer to the model

        Args:
            points (List[Point]): list of Point classes, in clockwise order (non closed simple polygon)
            soil_code (str): code of the soil for this layer
            label (str): label defaults to empty string
            notes (str): notes defaults to empty string
            stage_id (int): stage to add to, defaults to 0

        Returns:
            int: id of the added layer
        """
        stage_id = stage_id if stage_id else self.current_stage

        if not self.datastructure.has_stage(stage_id):
            raise IndexError(f"stage {stage_id} is not available")

        geometry = self.datastructure.geometries[stage_id]
        soillayerscollection = self.datastructure.soillayers[stage_id]

        # do we have this soil code?
        if not self.soils.has_soilcode(soil_code):
            raise ValueError(
                f"The soil with code {soil_code} is not defined in the soil collection."
            )

        # add the layer to the geometry
        # the checks on the validity of the points are done in the PersistableLayer class
        persistable_layer = geometry.add_layer(
            id=str(self._get_next_id()), label=label, points=points, notes=notes
        )

        # add the connection between the layer and the soil to soillayers
        soil = self.soils.get_soil(soil_code)
        soillayerscollection.add_soillayer(layer_id=persistable_layer.Id, soil_id=soil.id)
        return int(persistable_layer.Id)

    def add_head_line(
        self,
        points: List[Point],
        label: str = "",
        notes: str = "",
        is_phreatic_line: bool = False,
        stage_id: int = None,
    ) -> int:
        """
        Add head line to the model

        Args:
            points (List[Point]): list of Point classes
            label (str): label defaults to empty string
            notes (str): notes defaults to empty string
            is_phreatic_line (bool): set as phreatic line, defaults to False
            stage_id (int): stage to add to, defaults to current stage

        Returns:
            bool: id of the added headline
        """
        stage_id = stage_id if stage_id else self.current_stage

        if not self.datastructure.has_stage(stage_id):
            raise IndexError(f"stage {stage_id} is not available")

        waternet = self.waternets[stage_id]

        persistable_headline = waternet.add_head_line(
            str(self._get_next_id()), label, notes, points, is_phreatic_line
        )
        return int(persistable_headline.Id)

    def add_reference_line(
        self,
        points: List[Point],
        bottom_headline_id: int,
        top_head_line_id: int,
        label: str = "",
        notes: str = "",
        stage_id: int = None,
    ) -> int:
        """
        Add reference line to the model

        Args:
            points (List[Point]): list of Point classes
            bottom_headline_id (int): id of the headline to use as the bottom headline
            top_head_line_id (int): id of the headline to use as the top headline
            label (str): label defaults to empty string
            notes (str): notes defaults to empty string
            stage_id (int): stage to add to, defaults to 0

        Returns:
            int: id of the added reference line
        """
        stage_id = stage_id if stage_id else self.current_stage

        if not self.datastructure.has_stage(stage_id):
            raise IndexError(f"stage {stage_id} is not available")

        waternet = self.waternets[stage_id]

        persistable_referenceline = waternet.add_reference_line(
            str(self._get_next_id()),
            label,
            notes,
            points,
            str(bottom_headline_id),
            str(top_head_line_id),
        )
        return int(persistable_referenceline.Id)

    def add_state_point(
        self,
        state_point: DStabilityStatePoint,
        stage_id: int = None,
    ) -> int:
        """
        Add state point to the model

        Args:
            state_point (DStabilityStatePoint): DStabilityStatePoint class
            stage_id (int): stage_id (int): stage to add to, defaults to the current stage

        Returns:
            int: id of the added add_state_point

        Todo:
            Check if point lies within the given layer
        """
        stage_id = stage_id if stage_id else self.current_stage

        if not self.datastructure.has_stage(stage_id):
            raise IndexError(f"stage {stage_id} is not available")

        states = self.datastructure.states[stage_id]

        # check if the given layer id is valid
        try:
            _ = self.datastructure.geometries[stage_id].get_layer(state_point.layer_id)
        except ValueError:
            raise ValueError(f"No layer with id '{state_point.layer_id} in this geometry")

        # todo > check if point is in layer

        state_point.id = (
            self._get_next_id()
        )  # the user does not know the id so we have to add it
        persistable_statepoint = state_point._to_internal_datastructure()
        states.add_state_point(persistable_statepoint)
        return int(persistable_statepoint.Id)

    def add_state_line(
        self,
        points: List[Point],
        state_points: List[DStabilityStateLinePoint],
        stage_id: int = None,
    ) -> None:
        """
        Add state line. From the Soils, only the state parameters are used.

        points are a list of points with x,z
        state_point are a list of DStabilit.. where ONLY the x is used, the Z will be calculated

        No result PersistableStateLine has no Id
        """
        # 1. check of de gegeven points aanwezig zijn in de layers want ze MOETEN op een
        #    punt uit een layer liggen
        # 2. voeg deze punten toe aan de PersistableStateLine.Points
        # 3. voeg op alle x coordinaten in DStabilityStateLinePoint[] een PersistableStateLinePoint object toe
        #    let op dat points.xmin <= x <= points.xmax
        # 4. voeg de PersistableStateLine toe aan de interne datastructuur
        # 5. geef de id terug
        stage_id = stage_id if stage_id else self.current_stage

        if not self.datastructure.has_stage(stage_id):
            raise IndexError(f"stage {stage_id} is not available")

        states = self.datastructure.states[stage_id]

        # each point should belong to a layer
        persistable_points = []

        for point in points:
            point.id = self._get_next_id()  # assign a new id
            persistable_points.append(PersistablePoint(X=point.x, Z=point.z))

        persistable_state_line_points = []
        for state_point in state_points:
            state_point.id = self._get_next_id()  # assign a new id
            persistable_state_line_points.append(state_point._to_internal_datastructure())

        states.add_state_line(persistable_points, persistable_state_line_points)

    def add_load(
        self,
        load: DStabilityLoad,
        consolidations: Optional[List[Consolidation]] = None,
        stage_id: Optional[int] = None,
    ) -> None:
        """Add a load to the object.

        The geometry should be defined before adding loads.

        If no consolidations are provided, a Consolidation with default values will be made for each SoilLayer.
        It is not possible to set consolidation degrees of loads afterwards since they don't have an id.

        Args:
            load: A subclass of DStabilityLoad.
            stage_id: Id used to identify the stage to which the load is linked. If no stage_id is proved, the current stage_id will be taken.

        Raises:
            ValueError: When the provided load is no subclass of DStabilityLoad, an invalid stage_id is provided, or the datastructure is no longer valid.
        """
        stage_id = stage_id if stage_id is not None else self.current_stage

        if not issubclass(type(load), DStabilityLoad):
            raise ValueError(
                f"load should be a subclass of DstabilityReinforcement, received {load}"
            )
        if self.datastructure.has_soil_layers(stage_id) and self.datastructure.has_loads(
            stage_id
        ):
            if consolidations is None:
                consolidations = self._get_default_consolidations(stage_id)
            else:
                self._verify_consolidations(consolidations, stage_id)
            self.datastructure.loads[stage_id].add_load(load, consolidations)
        else:
            raise ValueError(f"No loads found for stage id {stage_id}")

    def add_soil_layer_consolidations(
        self,
        soil_layer_id: int,
        consolidations: Optional[List[Consolidation]] = None,
        stage_id: int = None,
    ) -> None:
        """Add consolidations for a layer (layerload).

        Consolidations cannot be added when adding soil layers since in the consolidations, all other soil layers need to be referred.
        Therefore, all soillayers in a stage should be defined before setting consolidation and
        the number of consolidations given should equal the amount of layers.

        Args:
            soil_layer_id: Consolidation is set for this soil layer id.
            consolidations: List of Consolidation. Must contain a Consolidation for every other layer.
            stage_id: Id used to identify the stage to which the load is linked. If no stage_id is proved, the current stage_id will be taken.

        Raises:
            ValueError: When the provided load is no subclass of DStabilityLoad, an invalid stage_id is provided, or the datastructure is no longer valid.
        """
        stage_id = stage_id if stage_id is not None else self.current_stage

        if self.datastructure.has_soil_layer(
            stage_id, soil_layer_id
        ) and self.datastructure.has_loads(stage_id):
            if consolidations is None:
                consolidations = self._get_default_consolidations(stage_id, soil_layer_id)
            else:
                self._verify_consolidations(consolidations, stage_id, soil_layer_id)

            self.datastructure.loads[stage_id].add_layer_load(
                soil_layer_id, consolidations
            )
        else:
            raise ValueError(f"No soil layers found found for stage id {stage_id}")

    def _get_default_consolidations(
        self, stage_id: int, exclude_soil_layer_id: Optional[int] = None
    ) -> List[Consolidation]:
        """Length of the consolidations is equal to the amount of soil layers.

        If exclude_soil_layer_id is provided, that specific soil layer id is not included in the consolidations.
        """
        if self.datastructure.has_soil_layers(stage_id):
            soil_layer_ids = self.datastructure.soillayers[stage_id].get_ids(
                exclude_soil_layer_id
            )
            return [Consolidation(layer_id=layer_id) for layer_id in soil_layer_ids]

        raise ValueError(f"No soil layers found for stage id {stage_id}")

    def _verify_consolidations(
        self,
        consolidations: List[Consolidation],
        stage_id: int,
        exclude_soil_layer_id: Optional[int] = None,
    ) -> None:
        if self.datastructure.has_soil_layers(stage_id):
            consolidation_soil_layer_ids: Set[str] = {
                str(c.layer_id) for c in consolidations
            }
            soil_layer_ids = self.datastructure.soillayers[stage_id].get_ids(
                exclude_soil_layer_id
            )

            if consolidation_soil_layer_ids != soil_layer_ids:
                raise ValueError(
                    f"Received consolidations ({consolidation_soil_layer_ids}) should contain all soil layer ids ({soil_layer_ids})"
                )
        else:
            raise ValueError(f"No soil layers found for stage id {stage_id}")

    def add_reinforcement(
        self,
        reinforcement: DStabilityReinforcement,
        stage_id: Optional[int] = None,
    ) -> None:
        """Add a reinforcement to the model.

        Args:
            reinforcement: A subclass of DStabilityReinforcement.
            stage_id: Id used to identify the stage to which the reinforcement is linked.

        Returns:
            int: Assigned id of the reinforcements (collection object of all reinforcements for a stage).

        Raises:
            ValueError: When the provided reinforcement is no subclass of DStabilityReinforcement, an invalid stage_id is provided, or the datastructure is no longer valid.
        """
        stage_id = stage_id if stage_id is not None else self.current_stage

        if not issubclass(type(reinforcement), DStabilityReinforcement):
            raise ValueError(
                f"reinforcement should be a subclass of DstabilityReinforcement, received {reinforcement}"
            )

        if self.datastructure.has_reinforcements(stage_id):
            self.datastructure.reinforcements[stage_id].add_reinforcement(reinforcement)
        else:
            raise ValueError(
                f"No reinforcements found for stage found with id {stage_id}"
            )

    def set_model(self, analysis_method: DStabilityAnalysisMethod, stage_id=None):
        """Sets the model and applies the given parameters

        Args:
            analysis_method (DStabilityAnalysisMethod): A subclass of DStabilityAnalysisMethod.
            stage_id: Id used to identify the stage to which the analysis method is linked.

        Raises:
            ValueError: When the provided analysismethod is no subclass of DStabilityAnalysisMethod,
            an invalid stage_id is provided, the analysis method is not known or the datastructure is no longer valid.
        """
        stage_id = stage_id if stage_id else self.current_stage

        if not self.datastructure.has_stage(stage_id):
            raise IndexError(f"stage {stage_id} is not available")

        calculationsettings = self.datastructure.calculationsettings[stage_id]

        _analysis_method_mapping = {
            AnalysisType.BISHOP: calculationsettings.set_bishop,
            AnalysisType.BISHOP_BRUTE_FORCE: calculationsettings.set_bishop_brute_force,
            AnalysisType.SPENCER: calculationsettings.set_spencer,
            AnalysisType.SPENCER_GENETIC: calculationsettings.set_spencer_genetic,
            AnalysisType.UPLIFT_VAN: calculationsettings.set_uplift_van,
            AnalysisType.UPLIFT_VAN_PARTICLE_SWARM: calculationsettings.set_uplift_van_particle_swarm,
        }

        try:
            _analysis_method_mapping[analysis_method.analysis_type](
                analysis_method._to_internal_datastructure()
            )
        except KeyError:
            raise ValueError(
                f"Unknown analysis method {analysis_method.analysis_type.value} found"
            )



from struct import pack, unpack_from

from pydantic.color import Color as PydanticColor


class Color(PydanticColor):
    """Color type from Pydantic

    see https://pydantic-docs.helpmanual.io/usage/types/#color-type

    Example:
        For construction::

            Color("ff00ff")  # hex value
            Color((255, 255, 255))  # rgb tuple
            Color("purple")  # named
    """

    def to_internal(self):
        r, g, b = self.as_rgb_tuple(alpha=False)
        return int.from_bytes(pack("BBBB", r, g, b, 0), "little")

    @classmethod
    def from_internal(cls, c: int):
        """Convert a D-Serie color integer into a Color."""
        r, g, b, _ = unpack_from("BBBB", c.to_bytes(4, "little"))
        return cls((r, g, b))



from typing import Any, Dict

from pydantic import FilePath

from geolib.models import BaseDataClass


class BaseSerializer(BaseDataClass):
    """Basic class for serializers."""

    ds: Dict[str, Any]

    def render(self):
        return str(self.ds)

    def write(self, filename: FilePath):
        """Test."""
        with open(filename, "w") as io:
            io.write(self.render())



from __future__ import annotations

import logging
from typing import Any, Dict, List, Tuple

from geolib.models.dseries_parser import (
    DSerieListGroupNextStructure,
    DSerieParser,
    DSeriesInlineProperties,
    DSeriesRepeatedGroupedProperties,
    DSeriesStructure,
    get_line_property_key_value,
    get_line_property_value,
    make_key,
    split_line_elements,
    strip_line_first_element,
)

logger = logging.getLogger(__name__)


class ComplexVerticalSubstructure(DSeriesStructure):
    """Specific Vertical substructure with a dedicated
    parser because of the oddities stored in that group."""

    def __init__(self, *args, **kwargs):
        """Replace unnamed arguments by fieldnames in order.

        Some structure in DSerie files have no names attached
        and are just values. This provides a way to still parse
        them.
        """
        largs = list(args)
        for field, fieldtype in self.__fields__.items():
            if len(largs) == 0:
                break
            if field in kwargs:
                continue
            else:
                value = strip_line_first_element(largs.pop(0))
                logger.debug(f"Setting {field}: {fieldtype} to {value}")
                kwargs[field] = value

        if len(largs) > 0:
            raise Exception(f"Failed to convert args to kwargs: {largs} left unmapped.")

        super().__init__(*args, **kwargs)

    @classmethod
    def parse_text(cls, text: str):
        """Parser that:

        - Knows about structure with fields and subgroups
        - Knows about repeated Vertical fields without a specific collection
        - Knows about the bug of time-dependent data groups that end without the - included
        - Knows about the bug of absence of the second end-tag of time-dependent data

        Arguments:
            data {str} -- Data to parse.
        """
        EXCEPTIONS = ["soil", "vertical", "time__dependent_data"]
        parsed_dict = {}
        currentkey = ""
        data = []
        args = []
        text_lines = list(filter(None, text.split("\n")))
        for i, line in enumerate(text_lines):
            sline = line.strip()
            # keyline
            if sline.startswith("[") and sline.endswith("]"):
                # [ key name ] => key_name
                key = make_key(sline[1:-1])

                # new group
                if currentkey == "":
                    logger.debug(f"Found new {key}")
                    currentkey = key
                    if len(data) > 0:
                        args.extend(data)
                    data = []

                # duplicate group before end
                elif currentkey == key:

                    # If key already exists, this is a group -> List
                    if currentkey in parsed_dict:
                        parsed_dict[currentkey] = parsed_dict[currentkey] + [
                            "".join(data)
                        ]
                    elif currentkey in EXCEPTIONS:
                        parsed_dict[currentkey] = ["".join(data)]
                    else:
                        parsed_dict[currentkey] = "".join(data)

                    data = []

                    logger.debug(
                        f"Duplicate key {key} at line {i} without first encountered and END OF."
                    )

                # end of current group
                elif key == "end_of_" + currentkey:
                    logger.debug(f"Found {key}")
                    # If key already exists, this is a group -> List
                    if currentkey in parsed_dict:
                        parsed_dict[currentkey] = parsed_dict[currentkey] + [
                            "".join(data)
                        ]
                    elif currentkey in EXCEPTIONS:
                        parsed_dict[currentkey] = ["".join(data)]
                    else:
                        parsed_dict[currentkey] = "".join(data)
                    data = ""
                    currentkey = ""

                # sub group that is eaten for now
                else:
                    data.append(line + "\n")

            # dataline
            else:
                data.append(line + "\n")

        # Because no [end] exist for duplicate group...
        parsed_dict[currentkey] = parsed_dict[currentkey] + ["".join(data)]

        return cls(*args, **parsed_dict)


class DSerieRepeatedTableStructure(DSeriesRepeatedGroupedProperties):
    @classmethod
    def group_value_is_list(cls, group_key: str) -> bool:
        """Override of parent method. Only the columns are not a list.

        Args:
            group_key (str): Property name.

        Returns:
            bool: Whether the key represents a list of values or not.
        """
        return group_key != "column_indication"

    @classmethod
    def get_validated_mappings(cls, generated_dict: Dict[str, str]) -> Dict[str, str]:
        """Transforms the generated dict from the parent into our own mapping.

        Args:
            generated_dict (Dict[str, str]): Parsed dictionary from parent.

        Returns:
            Dict[str, str]: Mapped dictionary with concrete properties.
        """
        class_name = cls.__name__.lower()
        n_groups = len(generated_dict.items())
        if n_groups != 2:
            raise ValueError(f"Expected 2 groups for {class_name} but got {n_groups}.")
        columns = [
            make_key(line)
            for line in generated_dict.pop("column_indication").split("\n")
            if line != ""
        ]
        parsed_dict: Dict[str, str] = {}
        # We assume there is only one 'repeated' key, so we get the values from it.
        for group_data in list(generated_dict.values())[0]:
            mapped_group = cls.get_mapped_group(columns, group_data)
            group_key = list(mapped_group.keys())[0]
            if group_key in parsed_dict.keys():
                raise ValueError(
                    f"No repeated table keys ({group_key}) allowed for {class_name}."
                )
            parsed_dict.update(mapped_group)

        return {cls.__name__.lower(): parsed_dict}

    @staticmethod
    def get_mapped_group(
        column_list: List[str], group_line: str
    ) -> Dict[str, List[Dict[str, str]]]:
        """Based on the input column and the group line, generates a list of
        dictionary entries which are mapped to the first element of the group
        line.

        Args:
            column_list (List[str]): List of columns to map.
            group_line (str): Content containing id of dictionary and list values.

        Returns:
            Dict[str, List[Dict[str, str]]]: Generated mapped group.
        """
        lines = [
            split_line_elements(line) for line in group_line.split("\n") if line != ""
        ]
        entry_id = lines.pop(0)[0]
        return {entry_id: [dict(zip(column_list, parts)) for parts in lines]}



from enum import Enum, IntEnum

from pydantic import conint

from geolib.models import BaseDataClass


class LateralEarthPressureMethod(IntEnum):
    """The method for input of the lateral earth pressure ratio"""

    MIXED = 2
    KA_KO_KP = 0
    C_PHI_DELTA = 1


class LateralEarthPressureMethodStage(IntEnum):
    """The method for input of the lateral earth pressure ratio"""

    KA_KO_KP = 1
    C_PHI_DELTA = 2

    @staticmethod
    def get_stage_type_from_method(
        method_value: LateralEarthPressureMethod,
    ):
        pair_dictionary = {
            LateralEarthPressureMethod.KA_KO_KP: LateralEarthPressureMethodStage.KA_KO_KP,
            LateralEarthPressureMethod.C_PHI_DELTA: LateralEarthPressureMethodStage.C_PHI_DELTA,
        }
        return pair_dictionary.get(method_value)


class ModelType(IntEnum):
    """Represents the model types from D-Sheet Piling using the same integer values used in the application.

    The Model type determines the kind of calculation that's performed on the construction.
    """

    SHEET_PILING = 0
    SINGLE_PILE = 1
    DIAPHRAGM_WALL = 2


class SinglePileLoadOptions(Enum):
    """Load options for the Single Pile"""

    LOADED_BY_FORCES = "forces"
    LOADED_BY_USER_DEFINED_DISPLACEMENTS = "user_defined_displacements"
    LOADED_BY_CALCULATED_DISPLACEMENTS = "calculated_displacements"


class PassiveSide(IntEnum):
    """Class to determine passive side."""

    DSHEETPILING_DETERMINED = 0
    LEFT = 1
    RIGHT = 2


class DistributionType(IntEnum):
    """Distribution type for probability analysis"""

    NONE = 0
    UNIFORM = 1
    NORMAL = 2
    LOG_NORMAL = 3
    EXPONENTIAL = 4


class Side(IntEnum):
    """Defines the two sides of a wall"""

    LEFT = 1
    RIGHT = 2
    BOTH = 3


class SheetPilingElementMaterialType(IntEnum):
    """Materials settings for sheet pile elements"""

    UserDefined = 0
    Steel = 1
    Concrete = 2
    # This is not implemented in the UI but it is possible input in the D-Series code
    Wood = 3
    Synthetic = 4
    Combined = 5


class CalculationType(IntEnum):
    """Calculation type for D-SheetPiling"""

    STANDARD = 0
    DESIGN_SHEETPILING_LENGTH = 1
    VERIFY_SHEETPILING = 2
    CHARACTERISTIC_KRANZ_ANCHOR_STRENGTH = 3
    OVERALL_STABILITY = 4
    RELIABILITY_ANALYSIS = 5


class DesignType(IntEnum):
    REPRESENTATIVE = 0
    CUR = 1
    EC7GENERAL = 2
    EC7NL = 3
    EC7BE = 4
    UNKNOWN = 5  # option is not available for the latest version of D-Sheet piling


class PartialFactorSetEC(IntEnum):
    DA1SET1 = 0
    DA1SET2 = 1
    DA2 = 2
    DA3 = 3


class PartialFactorSetEC7NADNL(IntEnum):
    RC0 = 0
    RC1 = 1
    RC2 = 2
    RC3 = 3


class PartialFactorSetEC7NADB(IntEnum):
    SET1 = 0
    SET2 = 1


class PartialFactorSetCUR(IntEnum):
    UNKNOWN = -1
    CLASSI = 0
    CLASSII = 1
    CLASSIII = 2


class PartialFactorCalculationType(IntEnum):
    METHODA = 0
    METHODB = 1


class VerifyType(IntEnum):
    CUR = 0
    EC7GENERAL = 1
    EC7NL = 2
    EC7BE = 3
    UNKNOWN = 4


class PartialFactorSetVerifyEC(IntEnum):
    DA1 = 0
    DA2 = 1
    DA3 = 2


class SoilTypeModulusSubgradeReaction(IntEnum):
    GRAVEL = 0
    SAND = 1
    LOAM = 2
    CLAY = 3
    PEAT = 4


class GrainType(IntEnum):
    FINE = 0
    COARSE = 1


class ModulusSubgradeReaction(IntEnum):
    MENARD = 0
    MANUAL = 1


class EarthPressureCoefficients(IntEnum):
    MANUAL = 0
    BRINCHHANSEN = 1


class LambdaType(IntEnum):
    MANUAL = 0
    MULLERBRESLAU = 1
    KOTTER = 2


class HorizontalBehaviorType(IntEnum):
    STIFF = 1
    ELASTIC = 2
    FOUNDATION = 3


class ModulusReactionType(Enum):
    """
    The Secant definition is based on the stress-displacement diagram according to CUR 166 of subgrade reaction.
    This diagram always uses three branches, with intersections at 50, 80 and 100 % of Ka−Kp
    The slope of the different branches is defined indirectly, via the three secant moduli at the intersection points.

    The Tangent (D-SheetPiling Classic) definition is based on a user-defined number of branches (number of curves),
    with the slope  of  each  branch  defined  directly  by  the  tangent  modulus
    """

    TANGENT = 0
    SECANT = 1


class CurveSettings(BaseDataClass):
    modulus_reaction_type: ModulusReactionType = ModulusReactionType.SECANT
    use_unloading_reloading_curve: bool = False
    curve_number: conint(ge=1, le=4) = 3


class LoadTypeFavourableUnfavourable(IntEnum):
    DSHEETPILING_DETERMINED = 0
    FAVOURABLE = 1
    UNFAVOURABLE = 2


class LoadTypeFavourableUnfavourableMoment(IntEnum):
    FAVOURABLE = 1
    UNFAVOURABLE = 2


class LoadTypePermanentVariable(IntEnum):
    PERMANENT = 0
    VARIABLE = 1



from abc import ABCMeta, abstractmethod
from datetime import timedelta
from typing import List, Optional, Union

from geolib.models import BaseDataClass

from .drain_types import DrainGridType, DrainSchedule, DrainType
from .internal import VerticalDrain as vd


class ScheduleValues(BaseDataClass, metaclass=ABCMeta):
    """ScheduleValues to inherit from"""

    schedule: Optional[DrainSchedule]

    @abstractmethod
    def _to_internal(self):
        """Creates internal datastructure from generic schedule values"""
        return


class ScheduleValuesOff(ScheduleValues):
    """Drainage Schedule with strips or columns is Off

    Arguments:
        Start of drainage -- The time t at which the drain becomes active. D-SETTLEMENT assumes
            that the water head in the drain equals the phreatic level.
        Phreatic level in drain -- The water head in the drain during drainage.
    """

    start_of_drainage: timedelta
    phreatic_level_in_drain: float

    def _to_internal(self, verticaldrains: vd) -> vd:
        verticaldrains.schedule_type = DrainSchedule.OFF
        verticaldrains.start_of_drainage = self.start_of_drainage.days
        verticaldrains.phreatic_level_in_drain = self.phreatic_level_in_drain
        return verticaldrains


class ScheduleValuesSimpleInput(ScheduleValues):
    """Drainage Schedule with strips or columns: Simple Input

    Arguments:
        Start of drainage -- The time t at which the drain becomes active. D-SETTLEMENT assumes
            that the water head in the drain equals the phreatic level.
        Begin time -- The time at which dewatering (i.e. a certain water level and air pressure) starts.
        End time -- The time at which dewatering stops.
        Underpressure -- The enforced underpressure during dewatering.
        Water head during dewatering -- The vertical level where the negative pore pressure equals the enforced underpressure during dewatering.
        Phreatic level in drain -- The water head in the drain during drainage.
    """

    start_of_drainage: timedelta
    phreatic_level_in_drain: float
    begin_time: float
    end_time: float
    underpressure: float
    water_head_during_dewatering: Optional[float]
    tube_pressure_during_dewatering: Optional[float]

    def _to_internal(self, verticaldrains: vd) -> vd:
        verticaldrains.schedule_type = DrainSchedule.SIMPLE_INPUT
        verticaldrains.start_of_drainage = self.start_of_drainage.days
        verticaldrains.phreatic_level_in_drain = self.phreatic_level_in_drain
        verticaldrains.begin_time = self.begin_time
        verticaldrains.end_time = self.end_time
        verticaldrains.water_head_during_dewatering = self.water_head_during_dewatering
        verticaldrains.tube_pressure_during_dewatering = (
            self.tube_pressure_during_dewatering
        )
        verticaldrains.under_pressure_for_strips_and_columns = self.underpressure
        verticaldrains.under_pressure_for_sand_wall = self.underpressure
        return verticaldrains


class ScheduleValuesDetailedInput(ScheduleValues):
    """Drainage Schedule with strips or columns: Detailed Input

    Arguments:
        Time -- The time at which dewatering is active.
        Underpressure -- This value is zero for vertical drains without enforced underpressure.
        Water head -- The vertical level where the negative pore pressure equals the enforced underpressure during dewatering.
    """

    time: List[timedelta] = []
    underpressure: List[float] = []
    water_level: List[float] = []

    def _to_internal(self, verticaldrains: vd) -> vd:
        verticaldrains.schedule_type = DrainSchedule.DETAILED_INPUT
        verticaldrains.time = [onetime.days for onetime in self.time]
        verticaldrains.underpressure = self.underpressure
        verticaldrains.water_level = self.water_level
        return verticaldrains


class VerticalDrain(BaseDataClass):
    """Vertical Drain Class to inherit from.
    This class refers to the input window "Vertical Drains" of the D-Settlement program.

    Arguments:
        range_from -- left limit to the drained area
        range_to -- right limit to the drained area
        bottom_position -- The (vertical) Y co-ordinate of the bottom end of the vertical drain.
        center_to_center -- The actual spacing between the drains.
        diameter -- The diameter of the Column drain.
        thickness -- The actual thickness of the Strip drain.
        grid -- The geometry of the grid.
    """

    drain_type: DrainType = DrainType.STRIP
    range_from: float
    range_to: float
    bottom_position: float
    center_to_center: float
    width: Optional[float]
    diameter: Optional[float]
    thickness: Optional[float]
    grid: DrainGridType = DrainGridType.UNDERDETERMINED
    schedule: ScheduleValues = None

    def _to_internal(self) -> vd:
        internal_vertical_drains = vd(
            drain_type=self.drain_type,
            range_from=self.range_from,
            range_to=self.range_to,
            bottom_position=self.bottom_position,
            center_to_center=self.center_to_center,
            width=self.width,
            diameter=self.diameter,
            thickness=self.thickness,
            grid=self.grid,
        )
        internal_vertical_drains = self.schedule._to_internal(internal_vertical_drains)
        return internal_vertical_drains



from pydantic import BaseModel


def children(instance, filter=BaseModel):
    """Recursively yield all children that are subclasses of filter."""
    for fieldname, field in instance:
        if isinstance(field, list):
            for f in field:
                if filter in f.__class__.__mro__:
                    yield f
                    yield from children(f)
                else:
                    print(f"Ignoring {f.__class__.__name__}")
        elif filter in field.__class__.__mro__:
            yield field
            yield from children(field)



"""
All D-Serie models store some metadata in the input files. These include
projectnames, dates and times, remarks etc. All of these options are
available via the `Metadata` class.

The `Metadata` class can also hold other properties for your project
such as an compute endpoint.

.. todo::
    Make a mapping between the possible metadata options and names for each model.

"""
from datetime import datetime
from pathlib import Path

from pydantic import AnyHttpUrl, BaseSettings, DirectoryPath

from geolib import __version__ as version

CONSOLE_RUN_BATCH_FLAG = "/b"


class MetaData(BaseSettings):
    """Holds all metadata found in the header of model files.

    Could be specified by default or in advance to make
    model generation easier.

    Also can read these settings automatically from a
    'geolib.env' file in the working directory, or give
    as '_env_file' parameter.
    """

    company: str = ""
    analyst: str = ""
    startdate: datetime = datetime.now()
    project: str = ""
    remarks: str = f"Created by GEOLib {version}"

    # For remote execution
    endpoint: AnyHttpUrl = "http://localhost:8000/"
    gl_username: str = "test"
    gl_password: str = "test"

    # For calculations
    console_folder: DirectoryPath = Path(".")
    timeout: int = 10 * 60  # in seconds, so 10 minutes

    # For multiple calculations
    calculation_folder: Path = Path("tests/test_output/calculations")
    nprocesses: int = 1

    # For ignoring extra fields that could come with newer/older versions
    # of input/output fields. We don't support any other value than "forbid"!
    extra_fields = "forbid"  # can be "ignore", "allow" or "forbid"

    class Config:
        env_file = "geolib.env"



"""
For all geometries and locations we try to use
a single Point class. 

In 1D applications it's only used to 
specify, for example height, other fields left None.

For profiles used in 1D applications, see :class:`~geolib.soils.layers.ProfileLayer`

"""
from math import isclose
from typing import Optional

from geolib.models import BaseDataClass

NODATA = -999.0  # TODO why is this implemented instead of None?


class Point(BaseDataClass):
    """A single Point Class."""

    label: Optional[str] = ""
    id: Optional[int]
    x: float = NODATA
    y: float = NODATA
    z: float = NODATA
    tolerance: float = 1e-4

    def __eq__(self, other):

        if isinstance(other, Point):
            return (
                isclose(self.x, other.x, abs_tol=self.tolerance)
                and isclose(self.y, other.y, abs_tol=self.tolerance)
                and isclose(self.z, other.z, abs_tol=self.tolerance)
            )
        else:
            return NotImplemented



from enum import IntEnum


class ProbabilisticCalculationType(IntEnum):
    SettlementsDeterministic = -1
    BandWidthOfSettlementsFOSM = 0
    ProbabilityOfFailureFORM = 1
    BandWidthAndProbabilityOfFailureMonteCarlo = 2



from datetime import datetime

from jinja2 import Environment, PackageLoader

from geolib import __version__ as glversion
from geolib.models.serializers import BaseSerializer

ENV = Environment(loader=PackageLoader("geolib.models.dfoundations"), trim_blocks=True)


class DFoundationsInputSerializer(BaseSerializer):
    def render(self) -> str:
        self.ds["input_data"].update(dict(timestamp=datetime.now()))
        self.ds["input_data"].update(dict(glversion=glversion))
        template = ENV.get_template("input.foi.j2")

        return template.render(self.ds["input_data"])



import secrets
import shutil
import uuid
from pathlib import Path, PosixPath, WindowsPath
from typing import Dict, List, Type, Union

import pydantic.json
from fastapi import BackgroundTasks, Depends, FastAPI, HTTPException
from fastapi.security import HTTPBasic, HTTPBasicCredentials
from pydantic import ValidationError, conlist
from starlette import status
from starlette.responses import JSONResponse

from geolib.errors import CalculationError
from geolib.models import (
    BaseModel,
    BaseModelList,
    DFoundationsModel,
    DSettlementModel,
    DSheetPilingModel,
    DStabilityModel,
)
from geolib.models.meta import MetaData

# Fixes for custom serialization
pydantic.json.ENCODERS_BY_TYPE[Path] = str
pydantic.json.ENCODERS_BY_TYPE[PosixPath] = str
pydantic.json.ENCODERS_BY_TYPE[WindowsPath] = str

settings = MetaData()
app = FastAPI()
security = HTTPBasic()


def get_current_username(credentials: HTTPBasicCredentials = Depends(security)):
    correct_username = secrets.compare_digest(credentials.username, settings.gl_username)
    correct_password = secrets.compare_digest(credentials.password, settings.gl_password)
    if not (correct_username and correct_password):
        raise HTTPException(
            status_code=status.HTTP_401_UNAUTHORIZED,
            detail="Incorrect email or password",
            headers={"WWW-Authenticate": "Basic"},
        )
    return credentials.username


@app.get("/users/me")
def read_current_user(username: str = Depends(get_current_username)):
    return {"username": username}


@app.get("/")
async def root():
    return {"message": "Hello World"}


def cleanup(path: Path):
    print(f"Cleaning up {path}")
    shutil.rmtree(path)


def execute(model, background_tasks: BackgroundTasks):
    unique_id = str(uuid.uuid4())
    unique_folder = Path(settings.calculation_folder / unique_id).absolute()
    unique_folder.mkdir(parents=True, exist_ok=True)
    ext = model.parser_provider_type().input_parsers[0].suffix_list[0]
    model.serialize(unique_folder / f"{unique_id}{ext}")

    # Override console folder from client
    model.meta.console_folder = settings.console_folder

    try:
        output = model.execute()
    except (CalculationError, ValidationError) as e:
        return JSONResponse(
            status_code=status.HTTP_500_INTERNAL_SERVER_ERROR,
            content={"message": str(e), "traceback": unique_id},
        )
    else:
        return output
    finally:
        background_tasks.add_task(cleanup, unique_folder)


@app.post("/calculate/dsettlementmodel")
async def calculate_dsettlementmodel(
    model: DSettlementModel,
    background_tasks: BackgroundTasks,
    _: str = Depends(get_current_username),
) -> DSettlementModel:
    return execute(model, background_tasks)


@app.post("/calculate/dfoundationsmodel")
async def calculate_dfoundationsmodel(
    model: DFoundationsModel,
    background_tasks: BackgroundTasks,
    _: str = Depends(get_current_username),
) -> DFoundationsModel:
    return execute(model, background_tasks)


@app.post("/calculate/dsheetpilingmodel")
async def calculate_dsheetpilingmodel(
    model: DSheetPilingModel,
    background_tasks: BackgroundTasks,
    _: str = Depends(get_current_username),
) -> DSheetPilingModel:
    return execute(model, background_tasks)


@app.post("/calculate/dstabilitymodel")
async def calculate_dstabilitymodel(
    model: DStabilityModel,
    background_tasks: BackgroundTasks,
    _: str = Depends(get_current_username),
) -> DStabilityModel:
    return execute(model, background_tasks)


@app.post("/calculate/dsettlementmodels")
async def calculate_many_dsettlementmodels(
    models: conlist(DSettlementModel, min_items=1),
    background_tasks: BackgroundTasks,
    _: str = Depends(get_current_username),
) -> List[DSettlementModel]:
    return execute_many(models, background_tasks)


@app.post("/calculate/dfoundationsmodels")
async def calculate_many_dfoundationsmodel(
    models: conlist(DFoundationsModel, min_items=1),
    background_tasks: BackgroundTasks,
    _: str = Depends(get_current_username),
) -> List[DFoundationsModel]:
    return execute_many(models, background_tasks)


@app.post("/calculate/dsheetpilingmodels")
async def calculate_many_dsheetpilingmodel(
    models: conlist(DSheetPilingModel, min_items=1),
    background_tasks: BackgroundTasks,
    _: str = Depends(get_current_username),
) -> List[DSheetPilingModel]:
    return execute_many(models, background_tasks)


@app.post("/calculate/dstabilitymodels")
async def calculate_many_dstabilitymodel(
    models: conlist(DStabilityModel, min_items=1),
    background_tasks: BackgroundTasks,
    _: str = Depends(get_current_username),
) -> List[DStabilityModel]:
    return execute_many(models, background_tasks)


def execute_many(
    models: BaseModelList, background_tasks: BackgroundTasks
) -> BaseModelList:
    unique_id = str(uuid.uuid4())
    unique_folder = Path(settings.calculation_folder / unique_id).absolute()
    unique_folder.mkdir(parents=True, exist_ok=True)

    bm = BaseModelList(models=models)

    try:
        output = bm.execute(unique_folder, nprocesses=settings.nprocesses)
    except (CalculationError, ValidationError) as e:
        return JSONResponse(
            status_code=status.HTTP_500_INTERNAL_SERVER_ERROR,
            content={"message": str(e), "traceback": unique_id},
        )
    else:
        return output
    finally:
        background_tasks.add_task(cleanup, unique_folder)



from .soil import (
    BjerrumParameters,
    ConeResistance,
    DistributionType,
    EarthPressureCoefficients,
    EarthPressureCoefficientsType,
    GrainType,
    HorizontalBehaviour,
    HorizontalBehaviourType,
    IsotacheParameters,
    KoppejanParameters,
    LambdaType,
    ModulusSubgradeReaction,
    MohrCoulombParameters,
    ShearStrengthModelTypePhreaticLevel,
    Soil,
    SoilClassificationParameters,
    SoilState,
    SoilStiffnessParameters,
    SoilType,
    SoilWeightParameters,
    StateLine,
    StatePoint,
    StateType,
    StochasticParameter,
    StorageParameters,
    StorageTypes,
    SubgradeReactionParameters,
    SuTablePoint,
    UndrainedParameters,
)



import abc
import re
from enum import Enum
from pathlib import Path
from typing import Dict, List, Optional, Set, Type, Union

from pydantic import DirectoryPath, FilePath

from geolib.geometry import Point
from geolib.models import BaseDataClass, BaseModel
from geolib.soils import Soil

from ...utils import camel_to_snake, snake_to_camel
from .dgeoflow_parserprovider import DGeoFlowParserProvider
from .internal import (
    CalculationType,
    CalculationTypeEnum,
    DGeoFlowResult,
    DGeoFlowStructure,
    GroundwaterFlowResult,
    PersistableSoilLayer,
    PipeLengthResult,
    PipeTrajectory,
    SoilCollection,
    SoilLayerCollection,
)
from .serializer import DGeoFlowInputSerializer, DGeoFlowInputZipSerializer


class DGeoFlowObject(BaseModel, metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def _to_DGeoFlow_sub_structure(self):
        raise NotImplementedError


class DGeoFlowModel(BaseModel):
    """D-Geoflow is software for soft soil piping calculations.

    This model can read, modify and create
    .flox files
    """

    current_scenario: int = -1
    current_scenario_index: int = 0
    current_calculation_index: int = 0
    datastructure: DGeoFlowStructure = DGeoFlowStructure()
    current_id: int = -1

    def __init__(self, *args, **data) -> None:
        super().__init__(*args, **data)
        self.current_id = self.datastructure.get_unique_id()

    @property
    def parser_provider_type(self) -> Type[DGeoFlowParserProvider]:
        return DGeoFlowParserProvider

    @property
    def console_path(self) -> Path:
        return Path("DGeoFlowConsole/DGeoFlow Console.exe")

    @property
    def console_flags_post(self) -> List[str]:
        return [
            str(self.current_scenario_index + 1),
            str(self.current_calculation_index + 1),
        ]

    @property
    def soils(self) -> SoilCollection:
        """Enables easy access to the soil in the internal dict-like datastructure. Also enables edit/delete for individual soils."""
        return self.datastructure.soils

    def get_layer(self, scenario_id: int, layer_id: int) -> PersistableSoilLayer:
        """Enables easy access to the soil in the internal dict-like datastructure. Also enables edit/delete for individual soils."""
        soillayers = self.datastructure.soillayers[scenario_id]
        for layer in soillayers.SoilLayers:
            if int(layer.LayerId) == layer_id:
                return layer
        raise ValueError(f"No soil layer found with layer id {layer_id}")

    def _get_next_id(self) -> int:
        self.current_id += 1
        return self.current_id

    def parse(self, *args, **kwargs):
        super().parse(*args, **kwargs)
        self.current_id = self.datastructure.get_unique_id()

    @property
    def output(self) -> DGeoFlowResult:
        return self.get_result(
            self.current_scenario_index, self.current_calculation_index
        )

    def get_result(self, scenario_index: int, calculation_index: int) -> Dict:
        """
        Returns the results of a scenario. Calculation results are based on analysis type and calculation type.

        Args:
            scenario_index (int): Index of a scenario.
            calculation_index (int): Index of a calculation.

        Returns:
            dict: Dictionary containing the analysis results of the scenario.

        Raises:
            ValueError: No results or calculationsettings available
        """
        return self._get_result_substructure(scenario_index, calculation_index)

    def _get_result_substructure(
        self, scenario_index: int, calculation_index: int
    ) -> DGeoFlowResult:
        if self.datastructure.has_result(scenario_index):
            calculation = self.datastructure.scenarios[scenario_index].Calculations[
                calculation_index
            ]
            result_id = calculation.ResultsId
            calculation_type = calculation.CalculationType

            results = self.datastructure.get_result_substructure(calculation_type)

            for result in results:
                if result.Id == result_id:
                    return result

        raise ValueError(f"No result found for result id {scenario_index}")

    def serialize(self, location: Union[FilePath, DirectoryPath]):
        """Support serializing to directory while developing for debugging purposes."""
        if not location.is_dir():
            serializer = DGeoFlowInputZipSerializer(ds=self.datastructure)
        else:
            serializer = DGeoFlowInputSerializer(ds=self.datastructure)
        serializer.write(location)
        self.filename = location

    def copy_scenario(self, label: str, notes: str, set_current=True) -> int:
        """Copy an existing scenario and add it to the model.

        Args:
            label: Label for the scenario
            notes: Notes for the scenario
            set_current: Whether to make the new scenario the current scenario.

        Returns:
            the id of the new scenario
        """
        new_id = self._get_next_id()
        new_scenario_id, new_unique_id = self.datastructure.duplicate_scenario(
            self.current_scenario, label, notes, new_id
        )

        if set_current:
            self.current_scenario = new_scenario_id
            self.current_scenario_index += 1
        self.current_id = new_unique_id
        return new_scenario_id

    def add_point(self, point: Point, scenario=None) -> int:
        """Add point, which should be unique in the model and return the created point id."""

    def add_soil(self, soil: Soil) -> int:
        """
        Add a new soil to the model. The code must be unique, the id will be generated

        Args:
            soil (Soil): a new soil

        Returns:
            int: id of the added soil
        """
        if self.soils.has_soilcode(soil.code):
            raise ValueError(f"The soil with code {soil.code} is already defined.")

        soil.id = self._get_next_id()
        persistant_soil = self.soils.add_soil(soil)
        return persistant_soil.Code

    def edit_soil(self, code: str, **kwargs: dict) -> None:
        """
        Edit an existing soil with parameter names based on the soil class members

        Args:
            code (str): the code of the soil
            kwargs (dict): the parameters and new values for example 'cohesion=2.0, friction_angel=25.0'

        Returns:
            bool: True for succes, False otherwise
        """
        return self.soils.edit_soil(code, **kwargs)

    @property
    def points(self):
        """Enables easy access to the points in the internal dict-like datastructure. Also enables edit/delete for individual points."""

    def add_layer(
        self,
        points: List[Point],
        soil_code: str,
        label: str = "",
        notes: str = "",
        scenario_id: int = None,
    ) -> int:
        """
        Add a soil layer to the model

        Args:
            points (List[Point]): list of Point classes, in clockwise order (non closed simple polygon)
            soil_code (str): code of the soil for this layer
            label (str): label defaults to empty string
            notes (str): notes defaults to empty string
            scenario_id (int): scenario to add to, defaults to 0

        Returns:
            int: id of the added layer
        """
        scenario_id = scenario_id if scenario_id else self.current_scenario

        if not self.datastructure.has_scenario(scenario_id):
            raise IndexError(f"scenario {scenario_id} is not available")
        geometry = self.datastructure.geometries[scenario_id]

        soillayerscollection = self.datastructure.soillayers[scenario_id]

        # do we have this soil code?
        if not self.soils.has_soilcode(soil_code):
            raise ValueError(
                f"The soil with code {soil_code} is not defined in the soil collection."
            )

        # add the layer to the geometry
        # the checks on the validity of the points are done in the PersistableLayer class
        persistable_layer = geometry.add_layer(
            id=str(self._get_next_id()), label=label, points=points, notes=notes
        )

        # add the connection between the layer and the soil to soillayers
        soil = self.soils.get_soil(soil_code)
        soillayerscollection.add_soillayer(layer_id=persistable_layer.Id, soil_id=soil.id)
        return int(persistable_layer.Id)

    def add_meshproperties(
        self,
        element_size: float = 1.0,
        label: str = "",
        scenario_id: int = None,
        layer_id: int = None,
    ) -> int:
        """
        Add a mesh properties to the model

        Args:
            element_size: size of the mesh elements for the discretization, defaults to 1.0
            scenario_id (int): scenario to add to, defaults to 0
            layer_id (int): layer to add to

        Returns:
            int: id of the added meshproperties collection
        """
        scenario_id = scenario_id if scenario_id else self.current_scenario
        meshpropertiescollection = self.datastructure.mesh_properties[scenario_id]

        persistable_layer = self.get_layer(scenario_id, layer_id)
        meshpropertiescollection.add_meshproperty(
            layer_id=persistable_layer.LayerId, element_size=element_size, label=label
        )

        return int(meshpropertiescollection.Id)

    def add_boundary_condition(
        self,
        points: List[Point],
        head_level: float,
        label: str = "",
        notes: str = "",
        scenario_id: int = None,
    ) -> int:
        """
        Add boundary conditions to the model

        Args:
            points (List[Point]): list of Point classes, in clockwise order (non closed simple polygon)
            head_level (float): level of the hydraulic head for the boundary condition
            label (str): label defaults to empty string
            notes (str): notes defaults to empty string
            scenario_id (int): scenario to add to, defaults to 0

        Returns:
            int: id of the boundary conditions collection
        """
        scenario_id = scenario_id if scenario_id else self.current_scenario
        boundary_conditions = self.datastructure.boundary_conditions[scenario_id]

        boundary_condition_id = self._get_next_id()
        boundary_conditions.add_boundary_condition(
            boundary_condition_id, label, notes, points, head_level
        )

        return boundary_condition_id

    def add_scenario(
        self,
        scenario_index: int = None,
        boundaryconditions_id: int = None,
        soillayers_id: int = None,
        geometry_id: int = None,
        meshproperties_id: int = None,
        label: str = "",
        notes: str = "",
        calculations_notes: str = "",
        stage_notes: str = "",
        calculations_label: str = None,
        stage_label: str = None,
    ) -> int:
        """
        Add a scenario to the model

        Args:
            scenario_index (int): scenario to add to, defaults to 0
            boundaryconditions_id (int): id of the boundary conditions collection to add to the scenario
            soillayers_id (int): id of the soil layers to add to the scenario
            geometry_id (int): id of the geometry to add to the scenario
            meshproperties_id (int): id of the mesh properties to add to the scenario
            label (str): label of the scenario, defaults to empty string
            notes (str): notes of the scenario, defaults to empty string
            calculations_notes (str): notes of the calculation, defaults to empty string
            stage_notes (str): notes of the stage, defaults to empty string
            calculations_label (str): label of the calculation, defaults to empty string
            stage_label (str): label of the stage, defaults to empty string

        Returns:
            int: id of the scenario
        """

        scenarios = self.datastructure.scenarios[scenario_index]

        scenarios.Label = label
        scenarios.Notes = notes
        scenarios.GeometryId = geometry_id
        scenarios.SoilLayersId = soillayers_id
        scenarios.add_calculation(
            label=calculations_label,
            notes=calculations_notes,
            mesh_properties_id=meshproperties_id,
        )
        scenarios.add_stage(
            label=stage_label,
            notes=stage_notes,
            boundaryconditions_collection_id=boundaryconditions_id,
        )
        return int(scenarios.Id)

    def set_calculation_type(
        self,
        scenario_index: int = 0,
        calculation_index: int = 0,
        calculation_type: CalculationTypeEnum = CalculationTypeEnum.GROUNDWATER_FLOW,
    ) -> None:
        """
        Sets the calculation type of a calculation.

        Args:
            scenario_index (int): The scenario index to add to, defaults to 0.
            calculation_index (int): The calculation index to add to, defaults to 0.
            calculation_type (CalculationTypeEnum): The calculation type, defaults to GROUNDWATER_FLOW.
        """

        self.datastructure.scenarios[scenario_index].Calculations[
            calculation_index
        ].CalculationType = calculation_type

    def set_pipe_trajectory(
        self,
        scenario_index: int = 0,
        calculation_index: int = 0,
        pipe_trajectory: PipeTrajectory = None,
    ) -> None:
        """
        Sets the pipe trajectory for a calculation.

        Args:
            scenario_index (int): The scenario index to add to, defaults to 0.
            calculation_index (int): The calculation index to add to, defaults to 0.
            pipe_trajectory (PipeTrajectory): The pipe trajectory.
        """

        self.datastructure.scenarios[scenario_index].Calculations[
            calculation_index
        ].PipeTrajectory = pipe_trajectory

    def set_critical_head_boundary_condition(
        self,
        scenario_index: int = 0,
        calculation_index: int = 0,
        boundary_condition_id: int = None,
    ) -> None:
        """
        Sets the critical head boundary condition for a calculation.

        Args:
            scenario_index (int): The scenario index to add to, defaults to 0.
            calculation_index (int): The calculation index to add to, defaults to 0.
            boundary_condition_id (int): The id of the critical head boundary condition.
        """

        self.datastructure.scenarios[scenario_index].Calculations[
            calculation_index
        ].CriticalHeadId = str(boundary_condition_id)

    def set_critical_head_search_parameters(
        self,
        scenario_index: int = 0,
        calculation_index: int = 0,
        minimum_head_level: float = 0,
        maximum_head_level: float = 1,
        step_size: float = 0.1,
    ) -> None:
        """
        Sets the critical head search parameters condition for a calculation.

        Args:
            scenario_index (int): The scenario index to add to, defaults to 0.
            calculation_index (int): The calculation index to add to, defaults to 0.
            minimum_head_level (float): The minimum head level to search at, defaults to 0.
            maximum_head_level (float): The maximum head level to search at, defaults to 1.
            step_size (float): The step size to search with, defaults to 0.1.
        """

        calculation = self.datastructure.scenarios[scenario_index].Calculations[
            calculation_index
        ]

        calculation.CriticalHeadSearchSpace.MinimumHeadLevel = minimum_head_level
        calculation.CriticalHeadSearchSpace.MaximumHeadLevel = maximum_head_level
        calculation.CriticalHeadSearchSpace.StepSize = step_size
